#!/usr/bin/python

##############################################

# Manoj M Wagle (USydney, MIT CSAIL)

##############################################


import os, sys, random, subprocess
os.environ["RETICULATE_PYTHON"] = sys.executable
import argparse
import logging
import glob
from captum.attr import *
from tqdm import tqdm
import pandas as pd
import numpy as np
import h5py
import pkg_resources

# Logging to Standard Error
Log_Format = "%(levelname)s - %(asctime)s - %(message)s \n"
logging.basicConfig(stream = sys.stderr, format = Log_Format, level = logging.INFO)

##############################################

# Check if the user requested help for the annotation script
if '--setting' in sys.argv and 'annotation' in sys.argv and ('--help' in sys.argv or '-h' in sys.argv):
    annotation_script_path = pkg_resources.resource_filename(__name__, 'Annotation.py')
    subprocess.run(["python", annotation_script_path, "--help"])
    sys.exit(0)

class CustomHelpFormatter(argparse.RawTextHelpFormatter):
    def format_help(self):
        help_text = super().format_help()
        welcome_message = (
            "\nThank you for using Hydra 😄, an interpretable deep generative tool for single-cell omics. Please refer to the full documentation available at https://sydneybiox.github.io/Hydra/ for detailed usage instructions. If you encounter any issues running the tool - Please open an issue on Github, and we will get back to you as soon as possible!!\n\n"
        )
        Note_message = "\n📍 NOTE 📍: You need to run feature selection (`fs`) on the train datatset before annotating the cell types in the query dataset. If you have already run feature selection on the train & want to annotate (`annotation`) a different related query dataset, please process the data (`processdata`) first and then provide the path to the directory containing this processed data.\n\n"

        return welcome_message + Note_message + help_text 

# Create argument parser
parser = argparse.ArgumentParser("Hydra", formatter_class=CustomHelpFormatter)

parser.add_argument('--seed', type = int, default = 42, help ='seed')

# Input
parser.add_argument('--train', help='Path to the training dataset (Seurat, SCE or Anndata object)')
parser.add_argument('--test', help='Path to the test dataset (Seurat or SCE object)')
parser.add_argument('--celltypecol', default='cell_type', help='Cell type label column in your input dataset (Seurat, SCE or Anndata object). Default: `cell_type`')
parser.add_argument('--modality', default='rna', choices=['rna', 'adt', 'atac'], help='Input data modality. Default: `rna`')
parser.add_argument('--base_dir', metavar = 'DIR', default=os.getcwd(), help = 'Path to the directory containing processed data directory. Default: Current working directory')
parser.add_argument('--gene', help='Name of the gene whose expression is to be highlighted in the plot')
parser.add_argument('--ctofinterest', help='Name of the cell type for which a ridgeline plot of gene expression should be generated')
parser.add_argument('--predictions', help='Generate UMAP plot for Hydra predicted cell types', default=False)
parser.add_argument('--ctpredictions', help='Path to the csv file containing cell types predicted by Hydra', default=False)
# parser.add_argument('--peak', help='If you are providing peak data for scATAC instead of Gene-activity, filtering will be turned off during data processing. This means that all peaks will be included', default=False)
parser.add_argument('--processdata_batch_size',  type = int, default = 1000, help = 'batch size for processing reference and query datasets')


# Training
parser.add_argument('--batch_size', type = int, default = 512, help = 'batch size for processing data during training')
parser.add_argument('--attr_batch_size', type = int, default = 500, help = 'batch size for feature atrribution. Please adjust this based on your GPU memory')
parser.add_argument('--epochs', type = int, default = 40, help = 'num of training epochs')
parser.add_argument('--lr', type = float, default = 0.02, help = 'learning rate')

# GPU specification    
parser.add_argument('--gpu', type = str, default = '0', help = 'Please specify the GPU to use')    

# Model
parser.add_argument('--z_dim', type = int, default = 100, help = 'Number of neurons in latent space')
parser.add_argument('--hidden_rna', type = int, default = 185, help = 'Number of neurons for RNA layer')
parser.add_argument('--hidden_adt', type = int, default = 30, help = 'Number of neurons for ADT layer')
parser.add_argument('--hidden_atac', type = int, default = 185, help = 'Number of neurons for ATAC layer')
parser.add_argument('--num_models', type = int, default = 25, help= 'Number of models for Ensemble Learning')

# Task 
parser.add_argument('--setting', type=str, required=True, 
    choices=['processdata', 'fs', 'plot', 'annotation'],
    help=(
        "`processdata` for processing input train and test Seurat, SCE or Anndata objects;\n"
        "`fs` for feature selection to obtain cell-identity genes;\n"
        "`plot` for generating UMAP plot of the dataset (Additionally, highlights gene expression when called with the `--gene` argument; Generates a ridgeline plot of expression of the specified gene in cell type of interest vs all other cell types when called with `--ctofinterest` argument; Generates a UMAP plot of Hydra predicted labels when called with `--predictions` argument);\n"
        "`annotation` for automated annotation of the query dataset\n\n"
    )
)

# Capture all remaining arguments for the annotation setting
parser.add_argument('annotation_args', nargs=argparse.REMAINDER, help='Additional arguments for annotation script')

# Parse the command-line arguments
args = parser.parse_args()

if args.gpu:
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu


##############################################
# Processing input data
def run_r_script(train_file, test_file, cell_type_label, data_type, processdata_batch_size):
    r_command = [
        "Rscript",
        pkg_resources.resource_filename(__name__, 'R/Process_Dataset.R'),
        train_file,
        test_file,
        cell_type_label,
        data_type,
        # str(peak),
        str(processdata_batch_size)
    ]
    try:
        subprocess.run(r_command, check=True)
    except subprocess.CalledProcessError:
        print("Error: The R script failed to execute.")
        sys.exit(1)


##############################################
def create_UMAP_plots(rds_file, modality, celltypecol, gene_name=None, ctofinterest=None):
    r_command = [
        "Rscript",
        pkg_resources.resource_filename(__name__, 'R/UMAP_plot.R'),
        rds_file,
        modality,
        celltypecol,
        gene_name if gene_name else "None",
        ctofinterest if ctofinterest else "None"
    ]
    try:
        subprocess.check_call(r_command)
    except subprocess.CalledProcessError as e:
        print(f"Error: The R script failed to execute. {e}")
        sys.exit(1)
    except FileNotFoundError as e:
        print(f"Error: File not found. {e}")
        sys.exit(1)


##############################################
def create_UMAP_Hydra_predictions(rds_file, modality, cell_type_predicted):
    r_command = [
        "Rscript",
        pkg_resources.resource_filename(__name__, 'R/plot_predictions.R'),
        rds_file,
        modality,
        cell_type_predicted
    ]
    try:
        subprocess.check_call(r_command)
    except subprocess.CalledProcessError as e:
        print(f"Error: The R script failed to execute. {e}")
        sys.exit(1)
    except FileNotFoundError as e:
        print(f"Error: File not found. {e}")
        sys.exit(1)


##############################################
# Query dataset annotation
def run_annotation_script(annotation_args):
    annotation_command = ["python", pkg_resources.resource_filename(__name__, 'Annotation.py')] + annotation_args
    subprocess.run(annotation_command, check=True)


##############################################

# Import torch-related libraries after setting the CUDA_VISIBLE_DEVICES
import torch
from torch.utils.data import DataLoader
import torch.nn as nn
from .model import (Autoencoder_CITEseq_Step1, Autoencoder_SHAREseq_Step1, Autoencoder_TEAseq_Step1,
                         Autoencoder_CITEseq_Step2, Autoencoder_SHAREseq_Step2, Autoencoder_TEAseq_Step2, 
                         Autoencoder_RNAseq_Step1, Autoencoder_RNAseq_Step2, Autoencoder_ADTseq_Step1,
                         Autoencoder_ADTseq_Step2, Autoencoder_ATACseq_Step1, Autoencoder_ATACseq_Step2)
from .train import train_model
from .util import (MyDataset, read_h5_data, Index2Label, read_fs_label,
                  load_and_preprocess_data, perform_data_augmentation, setup_seed)

# Check the device type based on GPU availability, MPS availability, or defaulting to CPU
device_str = "CUDA" if torch.cuda.is_available() \
            else "MPS" if torch.backends.mps.is_built() \
            else "CPU"
device = torch.device(device_str.lower())

FloatTensor = torch.FloatTensor
LongTensor = torch.LongTensor

setup_seed(args.seed) ### set seed for reproducbility


##############################################

print("\nThank you for using Hydra 😄, an interpretable deep generative tool for single-cell omics. Please refer to the full documentation available at https://sydneybiox.github.io/Hydra/ for detailed usage instructions. If you encounter any issues running the tool - Please open an issue on Github, and we will get back to you as soon as possible!!\n\n")

specified_gpus = args.gpu.split(',') if args.gpu else []
num_gpus_specified = len(specified_gpus)

# If CUDA_VISIBLE_DEVICES is not set, use PyTorch to get the total GPU count.
if not specified_gpus:
    num_gpus = torch.cuda.device_count()
else:
    num_gpus = num_gpus_specified

# Get the indices of GPUs that are currently visible to PyTorch
active_gpu_indices = os.environ.get("CUDA_VISIBLE_DEVICES", "").split(',')

if num_gpus >= 1 and (device_str == "CPU"):
    print("It seems the CPU version of PyTorch is installed. For GPU utilization, please install the GPU version of PyTorch. Currently, running on CPU!!!")
    
print("===============================\n")
print("Device to be used:", device_str, "\n")
print("===============================\n")


##############################################

def main():
    logging.info("Starting to run")

    if args.setting.lower() == 'processdata':
        if not args.train or not args.modality:
            parser.error("--train and --modality are required when --setting is 'processdata'")
        else:
            # Call the R script to process the data
            logging.info("Processing datasets...")
            test_file = args.test if args.test else "None"
            run_r_script(args.train, test_file, args.celltypecol, args.modality, args.processdata_batch_size)

    elif args.setting.lower() == 'plot':
        if args.predictions:
            if not args.test or not args.modality or not args.ctpredictions:
                parser.error("--test, --modality and --ctpredictions are required when --setting is 'plot' and --predictions is True")
            else:
                # Call the R script to create UMAP plots with predicted cell types
                logging.info("Generating plot for Hydra predicted cell types...")
                create_UMAP_Hydra_predictions(args.test, args.modality, args.ctpredictions)
        else:    
            if not args.train or not args.modality:
                parser.error("--train and --modality are required when --setting is 'UMAPplot'")
            else:
                # Call the R script to create UMAP plots
                logging.info("Generating plot...")
                create_UMAP_plots(args.train, args.modality, args.celltypecol, args.gene, args.ctofinterest)

    elif args.setting.lower() == 'fs':
        dataset_folder_path = os.path.join(args.base_dir, "Input_Processed")

        if not os.path.exists(dataset_folder_path):
            logging.info(f"Error: The dataset folder '{dataset_folder_path}' does not exist. Please provide path to the directory containing `Input_Processed` directory!!!")
            sys.exit(1)

        dataset_folders = sorted(glob.glob(f"{args.base_dir}/Input_Processed"))

        cwd = os.getcwd()

        for dataset_folder in dataset_folders:

            logging.info("Training model...")

            split_folders = sorted(glob.glob(f"{dataset_folder}/split_*"))

            for split_folder in split_folders:

                model_save_path = os.path.join(cwd, 'trained_model')
                os.makedirs(model_save_path, exist_ok = True)

                if os.path.isfile(f"{split_folder}/rna_train.h5"):
                    args.rna = f"{split_folder}/rna_train.h5"
                else:
                    args.rna = "NULL" 
                if os.path.isfile(f"{split_folder}/adt_train.h5"):
                    args.adt = f"{split_folder}/adt_train.h5"
                else:
                    args.adt = "NULL" 
                if os.path.isfile(f"{split_folder}/atac_train.h5"):
                    args.atac = f"{split_folder}/atac_train.h5"
                else:
                    args.atac = "NULL"
                args.cty = f"{split_folder}/ct_train.csv"


                if args.adt != "NULL" and args.atac != "NULL":
                    # Load and preprocess the data
                    (train_data, train_dl, train_label, mode, classify_dim, nfeatures_rna, nfeatures_adt, nfeatures_atac, feature_num, label_to_name_mapping) = load_and_preprocess_data(args, setting = "train")
                    logging.info("The Dataset is: scRNA+scADT+scATAC")

                if args.adt != "NULL" and args.atac == "NULL": 
                    # Load and preprocess the data
                    (train_data, train_dl, train_label, mode, classify_dim, nfeatures_rna, nfeatures_adt, feature_num, label_to_name_mapping) = load_and_preprocess_data(args, setting = "train")
                    logging.info("The Dataset is: scRNA+scADT")

                if args.adt == "NULL" and args.atac != "NULL":
                    # Load and preprocess the data
                    (train_data, train_dl, train_label, mode, classify_dim, nfeatures_rna, nfeatures_atac, feature_num, label_to_name_mapping) = load_and_preprocess_data(args, setting = "train")
                    logging.info("The Dataset is: scRNA+scATAC")
                                
                if args.adt == "NULL" and args.atac == "NULL": # scRNA-seq
                    # Load and preprocess the data
                    (train_data, train_dl, train_label, mode, classify_dim, nfeatures_rna, feature_num, label_to_name_mapping) = load_and_preprocess_data(args, setting = "train")
                    logging.info("The Dataset is: scRNA-seq")

                if args.atac == "NULL" and args.rna == "NULL": # scADT-Seq
                    # Load and preprocess the data
                    (train_data, train_dl, train_label, mode, classify_dim, nfeatures_adt, feature_num, label_to_name_mapping) = load_and_preprocess_data(args, setting = "train")
                    logging.info("The Dataset is: scADT-seq")

                if args.adt == "NULL" and args.rna == "NULL": # scATAC-Seq
                    # Load and preprocess the data
                    (train_data, train_dl, train_label, mode, classify_dim, nfeatures_atac, feature_num, label_to_name_mapping) = load_and_preprocess_data(args, setting = "train")
                    logging.info("The Dataset is: scATAC-seq")
                
                                
                ########## Step 1 ########### 

                ### Build model
                if mode == "scRNA+scADT+scATAC":
                    model = Autoencoder_TEAseq_Step1(nfeatures_rna, nfeatures_adt, nfeatures_atac, args.hidden_rna, args.hidden_adt, args.hidden_atac, args.z_dim, classify_dim, args.num_models)
                elif mode == "scRNA+scADT":
                    model = Autoencoder_CITEseq_Step1(nfeatures_rna, nfeatures_adt, args.hidden_rna, args.hidden_adt, args.z_dim, classify_dim, args.num_models)
                elif mode == "scRNA+scATAC":
                    model = Autoencoder_SHAREseq_Step1(nfeatures_rna, nfeatures_atac, args.hidden_rna, args.hidden_atac, args.z_dim, classify_dim, args.num_models)
                elif mode == "scRNA-seq":
                    model = Autoencoder_RNAseq_Step1(nfeatures_rna, args.hidden_rna, args.z_dim, classify_dim, args.num_models)
                elif mode == "scADT-seq":
                    model = Autoencoder_ADTseq_Step1(nfeatures_adt, args.hidden_adt, args.z_dim, classify_dim, args.num_models)
                elif mode == "scATAC-seq":
                    model = Autoencoder_ATACseq_Step1(nfeatures_atac, args.hidden_atac, args.z_dim, classify_dim, args.num_models)


                model = model.to(device)

                # Train the original model on the original raw data including all classifiers
                model, loss, train_num = train_model(model, train_dl, lr=args.lr, epochs=args.epochs, 
                        classify_dim=classify_dim, save_path=model_save_path, 
                        save_filename='Original_Model.pth.tar', feature_num=feature_num, 
                        use_balancing=True)

            
                checkpoint_tar = os.path.join(model_save_path, 'Original_Model.pth.tar')
                if os.path.exists(checkpoint_tar):
                    # Load the model's weights
                    checkpoint = torch.load(checkpoint_tar, weights_only=False)
                    model = model.module if isinstance(model, nn.DataParallel) else model
                    model.load_state_dict(checkpoint['state_dict'], strict=True)


                ########## Step 2 - Finetuning ###########

                min_epochs, max_epochs = 30, 50

                for modelI in range(args.num_models):

                    logging.info("\n\nRefining Model: %s", modelI+1)
                    
                    if mode == "scRNA+scADT+scATAC":
                        Step2_model = Autoencoder_TEAseq_Step2(nfeatures_rna, nfeatures_adt, nfeatures_atac, args.hidden_rna, args.hidden_adt, args.hidden_atac, args.z_dim, classify_dim)
                    elif mode == "scRNA+scADT":
                        Step2_model = Autoencoder_CITEseq_Step2(nfeatures_rna, nfeatures_adt, args.hidden_rna, args.hidden_adt, args.z_dim, classify_dim)
                    elif mode == "scRNA+scATAC":
                        Step2_model = Autoencoder_SHAREseq_Step2(nfeatures_rna, nfeatures_atac, args.hidden_rna, args.hidden_atac, args.z_dim, classify_dim)
                    elif mode == "scRNA-seq":
                        Step2_model = Autoencoder_RNAseq_Step2(nfeatures_rna, args.hidden_rna, args.z_dim, classify_dim)
                    elif mode == "scADT-seq":
                        Step2_model = Autoencoder_ADTseq_Step2(nfeatures_adt, args.hidden_adt, args.z_dim, classify_dim)
                    elif mode == "scATAC-seq":
                        Step2_model = Autoencoder_ATACseq_Step2(nfeatures_atac, args.hidden_atac, args.z_dim, classify_dim)

                    # Load the encoder weights trained from Step 1 
                    encoder_weights = model.encoder.state_dict() 
                    Step2_model.encoder.load_state_dict(encoder_weights)

                    # Load the decoder weights trained from Step 1 
                    decoder_weights = model.decoder.state_dict() 
                    Step2_model.decoder.load_state_dict(decoder_weights)

                    # Load the classifier weights trained from Step 1 
                    classifier_weights = model.classifiers[modelI].state_dict()
                    Step2_model.classify.load_state_dict(classifier_weights)

                    # Move the model to GPU and automatically utilize multiple GPUs if available
                    Step2_model = Step2_model.to(device)
                                                        
                    # Generate Augmented dataset  
                    new_data, new_label, new_label_names = perform_data_augmentation(label_to_name_mapping, train_num, classify_dim, train_label, train_data, model, args)

                    # Process the new data after augmentation
                    train_transformed_dataset = MyDataset(new_data, new_label)

                    new_train_dl = DataLoader(train_transformed_dataset, batch_size=args.batch_size, shuffle=True, num_workers=0, drop_last = False)

                    # set a random number of epochs for this model
                    epochs = random.randint(min_epochs, max_epochs)

                    Step2_model, loss, _ = train_model(Step2_model, new_train_dl, lr=args.lr, epochs=epochs, 
                                            classify_dim=classify_dim, save_path=model_save_path, 
                                            save_filename='FineTuned_Model.pth.tar', feature_num=feature_num, 
                                            use_balancing=False)
                    
                    # Create directory for balanced data if it does not exist
                    balanced_data_dir = os.path.join(cwd, f'Balanced_Data-{args.num_models}')
                    os.makedirs(balanced_data_dir, exist_ok = True)
                    
                    new_data, new_label, new_label_names = perform_data_augmentation(label_to_name_mapping, train_num, classify_dim, train_label, train_data, Step2_model, args)

                    # Modify new_data_path and new_label_path to use balanced_data_dir
                    new_data_path = os.path.join(balanced_data_dir, f'train_data_bal_{modelI+1}.pt')
                    torch.save(new_data, new_data_path)

                    new_label_path = os.path.join(balanced_data_dir, f'train_label_bal_{modelI+1}.pt')
                    torch.save(new_label, new_label_path)                                        
                
                    # Create directory for final models if it does not exist
                    final_models_dir = os.path.join(model_save_path, f'Final_Models-{args.num_models}')
                    os.makedirs(final_models_dir, exist_ok = True)

                    final_model = Step2_model.module if isinstance(Step2_model, nn.DataParallel) else Step2_model

                    # Save the final model
                    final_model_path = os.path.join(final_models_dir, f'model_{modelI+1}.pth.tar')
                    torch.save({'state_dict': final_model.state_dict()}, final_model_path)


                ############ Run feature selection ############  

                logging.info("\n\nRunning feature selection...")

                # Convert feature names to string format
                rna_name_new = []
                adt_name_new = []
                atac_name_new = []

                if os.path.isfile(f"{split_folder}/rna_train.h5"):
                    rna_data_path = f"{split_folder}/rna_train.h5"
                    rna_data_path_noscale = f"{split_folder}/rna_train_noscale.h5"
                else:
                    rna_data_path = "NULL" 
                if os.path.isfile(f"{split_folder}/adt_train.h5"):
                    adt_data_path = f"{split_folder}/adt_train.h5"
                    adt_data_path_noscale = f"{split_folder}/adt_train_noscale.h5"
                else:
                    adt_data_path = "NULL" 
                if os.path.isfile(f"{split_folder}/atac_train.h5"):
                    atac_data_path = f"{split_folder}/atac_train.h5"
                    atac_data_path_noscale = f"{split_folder}/atac_train_noscale.h5"
                else:
                    atac_data_path = "NULL"
                label_path = f"{split_folder}/ct_train.csv"

                (label, _) = read_fs_label(label_path)

                index_to_label = Index2Label(label_path, classify_dim)

                classify_dim = (max(label)+1).cpu().numpy()

                if adt_data_path != "NULL" and atac_data_path != "NULL" and rna_data_path != "NULL":
                    mode = "scRNA+scADT+scATAC"

                    rna_name = h5py.File(rna_data_path, "r")['matrix/features'][:]
                    adt_name = h5py.File(adt_data_path, "r")['matrix/features'][:]
                    atac_name = h5py.File(atac_data_path, "r")['matrix/features'][:]

                    rna_data = read_h5_data(rna_data_path)
                    adt_data = read_h5_data(adt_data_path)
                    atac_data = read_h5_data(atac_data_path)

                    rna_data_noscale = read_h5_data(rna_data_path_noscale)
                    adt_data_noscale = read_h5_data(adt_data_path_noscale)
                    atac_data_noscale = read_h5_data(atac_data_path_noscale)

                    nfeatures_rna = rna_data.shape[1]
                    nfeatures_adt = adt_data.shape[1]
                    nfeatures_atac = atac_data.shape[1]

                    feature_num = nfeatures_rna + nfeatures_adt + nfeatures_atac

                    data = torch.cat((rna_data, adt_data, atac_data), 1)
                    data_noscale = torch.cat((rna_data_noscale, adt_data_noscale, atac_data_noscale), 1)
                
                if adt_data_path != "NULL" and atac_data_path == "NULL":
                    mode = "scRNA+scADT"

                    rna_name = h5py.File(rna_data_path, "r")['matrix/features'][:]
                    adt_name = h5py.File(adt_data_path, "r")['matrix/features'][:]

                    rna_data = read_h5_data(rna_data_path)
                    adt_data = read_h5_data(adt_data_path)

                    rna_data_noscale = read_h5_data(rna_data_path_noscale)
                    adt_data_noscale = read_h5_data(adt_data_path_noscale)

                    nfeatures_rna = rna_data.shape[1]
                    nfeatures_adt = adt_data.shape[1]

                    feature_num = nfeatures_rna + nfeatures_adt

                    data = torch.cat((rna_data, adt_data), 1)
                    data_noscale = torch.cat((rna_data_noscale, adt_data_noscale), 1)
                
                if adt_data_path == "NULL" and atac_data_path != "NULL":
                    mode = "scRNA+scATAC"

                    rna_name = h5py.File(rna_data_path, "r")['matrix/features'][:]
                    atac_name = h5py.File(atac_data_path, "r")['matrix/features'][:]

                    rna_data = read_h5_data(rna_data_path)
                    atac_data = read_h5_data(atac_data_path)

                    rna_data_noscale = read_h5_data(rna_data_path_noscale)
                    atac_data_noscale = read_h5_data(atac_data_path_noscale)

                    nfeatures_rna = rna_data.shape[1]
                    nfeatures_atac = atac_data.shape[1]

                    feature_num = nfeatures_rna + nfeatures_atac

                    data = torch.cat((rna_data, atac_data), 1)
                    data_noscale = torch.cat((rna_data_noscale, atac_data_noscale), 1)
                
                if adt_data_path == "NULL" and atac_data_path == "NULL":
                    mode = "scRNA-seq"

                    rna_name = h5py.File(rna_data_path, "r")['matrix/features'][:]

                    rna_data = read_h5_data(rna_data_path)
                    rna_data_noscale = read_h5_data(rna_data_path_noscale)

                    nfeatures_rna = rna_data.shape[1]

                    feature_num = nfeatures_rna

                    data = rna_data
                    data_noscale = rna_data_noscale

                if rna_data_path == "NULL" and atac_data_path == "NULL":
                    mode = "scADT-seq"

                    adt_name = h5py.File(adt_data_path, "r")['matrix/features'][:]

                    adt_data = read_h5_data(adt_data_path)
                    adt_data_noscale = read_h5_data(adt_data_path_noscale)

                    nfeatures_adt = adt_data.shape[1]

                    feature_num = nfeatures_adt

                    data = adt_data
                    data_noscale = adt_data_noscale
                
                if rna_data_path == "NULL" and atac_data_path == "NULL":
                    mode = "scATAC-seq"

                    atac_name = h5py.File(atac_data_path, "r")['matrix/features'][:]

                    atac_data = read_h5_data(atac_data_path)
                    atac_data_noscale = read_h5_data(atac_data_path_noscale)

                    nfeatures_atac = atac_data.shape[1]

                    feature_num = nfeatures_atac

                    data = atac_data
                    data_noscale = atac_data_noscale


                if mode == "scRNA+scADT+scATAC":
                    for i in range(nfeatures_rna):
                        a = str(rna_name[i], encoding="utf-8") + "_RNA_"
                        rna_name_new.append(a)
                    for i in range(nfeatures_adt):
                        a = str(adt_name[i], encoding="utf-8") + "_ADT_"
                        adt_name_new.append(a)
                    for i in range(nfeatures_atac):
                        a = str(atac_name[i], encoding="utf-8") + "_ATAC_"
                        atac_name_new.append(a)
                    features = rna_name_new + adt_name_new + atac_name_new

                if mode == "scRNA+scADT":
                    for i in range(nfeatures_rna):
                        a = str(rna_name[i], encoding="utf-8") + "_RNA_"
                        rna_name_new.append(a)
                    for i in range(nfeatures_adt):
                        a = str(adt_name[i], encoding="utf-8") + "_ADT_"
                        adt_name_new.append(a)
                    features = rna_name_new + adt_name_new

                if mode == "scRNA+scATAC":
                    for i in range(nfeatures_rna):
                        a = str(rna_name[i], encoding="utf-8") + "_RNA_"
                        rna_name_new.append(a)
                    for i in range(nfeatures_atac):
                        a = str(atac_name[i], encoding="utf-8") + "_ATAC_"
                        atac_name_new.append(a)
                    features = rna_name_new + atac_name_new
                
                if mode == "scRNA-seq":
                    for i in range(nfeatures_rna):
                        a = str(rna_name[i], encoding="utf-8") + "_RNA_"
                        rna_name_new.append(a)
                    features = rna_name_new

                if mode == "scADT-seq":
                    for i in range(nfeatures_adt):
                        a = str(adt_name[i], encoding="utf-8") + "_ADT_"
                        adt_name_new.append(a)
                    features = adt_name_new
                
                if mode == "scATAC-seq":
                    for i in range(nfeatures_atac):
                        a = str(atac_name[i], encoding="utf-8") + "_ATAC_"
                        atac_name_new.append(a)
                    features = atac_name_new

                # Load all model files
                model_files = glob.glob(os.path.join(model_save_path, f'Final_Models-{args.num_models}/*.pth.tar'))

                # Run feature selection for each cell type
                for i in tqdm(range(classify_dim)):

                    cell_type_name = index_to_label[i]

                    cell_type_name = cell_type_name.replace("/", "_")

                    # Select the data for the current cell type and for all other cell types
                    current_type_data = data_noscale[torch.where(label == i)].reshape(-1, feature_num)
                    other_type_data = data_noscale[torch.where(label != i)].reshape(-1, feature_num)

                    # Calculate the mean expression for each feature across the two groups
                    mean_current_type = torch.mean(current_type_data, dim=0)
                    mean_other_types = torch.mean(other_type_data, dim=0)

                    # Compute fold changes for each feature
                    epsilon = 1e-6
                    fold_changes = (mean_current_type + epsilon) / (mean_other_types + epsilon)

                    # Apply log transformation to fold changes
                    log_fold_changes = torch.log2(fold_changes)

                    # Get indices of cells with the current cell type
                    train_index_fs = torch.where(label == i)
                    train_index_fs = [t.cpu().numpy() for t in train_index_fs]
                    train_index_fs = np.array(train_index_fs)

                    # Get data for the current cell type
                    train_data_each_celltype_fs = data[train_index_fs, :].reshape(-1, feature_num)

                    attributions_all_models = []

                    # Compute the attribution for each cell of the current cell type
                    for model_file in model_files:
                        if mode == "scRNA+scADT+scATAC":
                            model_test = Autoencoder_TEAseq_Step2(nfeatures_rna, nfeatures_adt, nfeatures_atac, args.hidden_rna, args.hidden_adt, args.hidden_atac, args.z_dim, classify_dim)
                        elif mode == "scRNA+scADT":
                            model_test = Autoencoder_CITEseq_Step2(nfeatures_rna, nfeatures_adt, args.hidden_rna, args.hidden_adt, args.z_dim, classify_dim)
                        elif mode == "scRNA+scATAC":
                            model_test = Autoencoder_SHAREseq_Step2(nfeatures_rna, nfeatures_atac, args.hidden_rna, args.hidden_atac, args.z_dim, classify_dim)
                        elif mode == "scRNA-seq":
                            model_test = Autoencoder_RNAseq_Step2(nfeatures_rna, args.hidden_rna, args.z_dim, classify_dim)
                        elif mode == "scADT-seq":
                            model_test = Autoencoder_ADTseq_Step2(nfeatures_adt, args.hidden_adt, args.z_dim, classify_dim)
                        elif mode == "scATAC-seq":
                            model_test = Autoencoder_ATACseq_Step2(nfeatures_atac, args.hidden_atac, args.z_dim, classify_dim)

                        # Load the model's weights
                        checkpoint = torch.load(model_file, weights_only=False)
                        model_test.load_state_dict(checkpoint['state_dict'], strict=True)

                        model_test = model_test.to(device)

                        classify_model = nn.Sequential(*list(model_test.children()))[0:2]

                        deconv = IntegratedGradients(classify_model)

                        Attr_batch_size = args.attr_batch_size

                        # Initialize the attributions tensor
                        attribution = torch.zeros(1, feature_num).to(device)

                        # Calculate the attributions in batches
                        for j in range(0, train_data_each_celltype_fs.size(0), Attr_batch_size):
                            batch = train_data_each_celltype_fs[j:j + Attr_batch_size, :]
                            batch = batch.to(device)
                            attribution += torch.sum(torch.abs(deconv.attribute(batch, target=i)), dim=0, keepdim=True)

                        # take mean attribution for the current model
                        attribution_mean = torch.mean(attribution, dim=0)
                        attributions_all_models.append(attribution_mean)

                        del model_test  # delete the current model to free up memory
                        torch.cuda.empty_cache()  # empty GPU cache to avoid out-of-memory errors

                    # calculate the average attribution across all models
                    average_attribution = sum(attributions_all_models) / len(model_files)

                    fs_score = average_attribution.reshape(-1).detach().cpu().numpy()

                    # Adjust by the sign of the log fold changes
                    fs_score = fs_score * np.sign(log_fold_changes.numpy())

                    # Directly sort by the scores themselves
                    indices_sorted = np.argsort(-fs_score)  # This sorts the scores in descending order

                    # Use the sorted indices to order your features and scores
                    fs_results = [features[index] + str(index) for index in indices_sorted]
                    fs_scores = [fs_score[index] for index in indices_sorted]

                    # Convert fs_results to a pandas DataFrame
                    fs_results_df = pd.DataFrame({'Feature Name': fs_results, 'Score': fs_scores})

                    # Saving feature selection results
                    cwd = os.getcwd()  # Get the current working directory
                    Hydra_folder = os.path.join(cwd, f"Results/Feature_Selection/Hydra-{args.num_models}")

                    if not os.path.exists(Hydra_folder):
                        os.makedirs(Hydra_folder)
                    # Save the feature selection results to a CSV file
                    fs_results_df.to_csv(os.path.join(Hydra_folder, f'fs.{cell_type_name}_Hydra.csv'), index=False)
    
    elif args.setting.lower() == "annotation":
        run_annotation_script(args.annotation_args)

    logging.info("Completed successfully!")

    return


##############################################

if __name__ == "__main__":

    # Call the main function
    main()



#!/usr/bin/python

##############################################

# Manoj M Wagle (USydney, MIT CSAIL)

##############################################


import os, sys
import argparse
import logging
import glob
from tqdm import tqdm
from captum.attr import *
import scipy.special

import pandas as pd
import numpy as np
import pickle

# Create argument parser
parser = argparse.ArgumentParser("Annotation")

# Add arguments to the parser
parser.add_argument('--seed', type = int, default = 42, help ='seed')

# Input Data
parser.add_argument('--proccessedata_dir', metavar = 'DIR', default = os.getcwd(), help = 'Path to the directory containing processed single-cell data. Default: Current working directory')  
parser.add_argument('--balanceddata_dir', metavar = 'DIR', default = os.getcwd(), help = 'Path to the directory containing balanced single-cell data. Default: Current working directory')
parser.add_argument('--fs_dir', metavar = 'DIR', default = os.getcwd(), help = 'Path to the directory containing feature selection results. Default: Current working directory')

# For training
parser.add_argument('--batch_sizeanno', type = int, default = 64, help = 'batch size')
parser.add_argument('--epochsanno', type = int, default = 5, help = 'num of training epochs')
parser.add_argument('--lranno', type = float, default = 0.01, help = 'init learning rate')

# GPU specification    
parser.add_argument('--gpus', type = str, default = '0', help = 'Please specify the GPU to use')   

# Model
parser.add_argument('--num_classifiers', type = int, default = 25, help= 'Number of classifiers - Should be same as the number of models used for feature selection')

# Parse the command-line arguments
args = parser.parse_args()

if args.gpus:
    os.environ["CUDA_VISIBLE_DEVICES"] = args.gpus


##############################################

# Import torch-related libraries after setting the CUDA_VISIBLE_DEVICES

import torch
from torch.utils.data import DataLoader
import torch.nn as nn
import torch.nn.functional as F
from torch.optim import Adam
from torch.utils.data import DataLoader, TensorDataset
from hydra.util import (setup_seed, real_label, NN_Classifier,
                  process_data_annotation_train, process_data_annotation_query)

device = torch.device("cuda" if torch.cuda.is_available()
                      else "mps" if getattr(torch, 'has_mps', False)
                      else "cpu")

print("Device to be used: ", device, "\n")

FloatTensor = torch.FloatTensor
LongTensor = torch.LongTensor

setup_seed(args.seed) ### set seed for reproducbility


##############################################

def main():

    dataset_folder_path = os.path.join(args.proccessedata_dir, "Input_Processed")

    if not os.path.exists(dataset_folder_path):
        logging.info(f"Error: The dataset folder '{dataset_folder_path}' does not exist. Please provide path to the directory containing `Input_Processed` directory!!!")
        sys.exit(1)

    dataset_folders = sorted(glob.glob(f"{args.proccessedata_dir}/Input_Processed"))

    for dataset_folder in dataset_folders:
        split_folders = sorted(glob.glob(f"{dataset_folder}/split_*"))

        for split_folder in split_folders:

            cwd = os.getcwd()
            feature_dir = f"{args.fs_dir}/Results/Feature_Selection/Hydra-{args.num_classifiers}/"

            args.rna = f"{split_folder}/rna_train.h5"
            if os.path.isfile(f"{split_folder}/adt_train.h5"):
                args.adt = f"{split_folder}/adt_train.h5"
            else:
                args.adt = "NULL" 
            if os.path.isfile(f"{split_folder}/atac_train.h5"):
                args.atac = f"{split_folder}/atac_train.h5"
            else:
                args.atac = "NULL"
            args.cty = f"{split_folder}/ct_train.csv"

            # Get a list of all balanced datasets in the split_folder
            balanced_data_files = sorted(glob.glob(f"{args.balanceddata_dir}/Balanced_Data-{args.num_classifiers}/train_data_bal_*.pt"))
            balanced_label_files = sorted(glob.glob(f"{args.balanceddata_dir}/Balanced_Data-{args.num_classifiers}/train_label_bal_*.pt"))

            model_save_dir = os.path.join(cwd, f"trained_model/Annotation_models-{args.num_classifiers}")
            os.makedirs(model_save_dir, exist_ok=True)  

            trainL = pd.read_csv(f"{split_folder}/ct_train.csv")

            save_path = os.path.join(model_save_dir, "ct_train.csv")

            trainL.to_csv(save_path, index=False)

            for idx, (data_file, label_file) in enumerate(zip(balanced_data_files, balanced_label_files)):

                # Train the classifer for the current balanced dataset
                logging.info(f"Training classifier: {idx+1}")

                args.balanced_data = data_file
                args.balanced_label = label_file
                
                (train_data, train_label, classify_dim) = process_data_annotation_train(args, feature_dir, split_folder)

                dataset = TensorDataset(train_data, train_label)
                dataloader = DataLoader(dataset, batch_size=args.batch_sizeanno, shuffle=True, drop_last = True)
                
                # Create NN classifier
                model = NN_Classifier(train_data.shape[1], classify_dim)

                if torch.cuda.device_count() > 1:
                    model = model.to(device)
                    model = nn.DataParallel(model)
                else:
                    model = model.to(device)

                criterion = nn.CrossEntropyLoss()
                optimizer = Adam(model.parameters(), args.lranno)

                # Training the model
                model.train()

                for epoch in tqdm(range(args.epochsanno)): 
                    epoch_loss = 0
                    correct_predictions = 0
                    total_predictions = 0

                    for batch in dataloader:
                        data, target = batch

                        data, target = data.to(device), target.to(device)

                        optimizer.zero_grad()

                        output = model(data)

                        loss = criterion(output, target)
                        epoch_loss += loss.item() 

                        loss.backward()

                        optimizer.step()

                        # Computing accuracy
                        predicted = torch.argmax(output, dim=1).to(device)
                        correct = (predicted == target).sum()  
                        correct_predictions += correct.item()
                        total_predictions += target.shape[0] 

                    epoch_loss /= len(dataloader)  # Taking average loss per epoch
                    accuracy = correct_predictions / total_predictions  # calculating accuracy

                    # print(f'Epoch {epoch+1}/{args.epochsanno}: Loss={epoch_loss}, Accuracy={accuracy}')

                # Save the model for this balanced dataset
                model_save_path = os.path.join(model_save_dir, f'FFNN_model_Hydra_{idx+1}.pt')
                final_model = model.module if isinstance(model, nn.DataParallel) else model
                torch.save(final_model.state_dict(), model_save_path)


            ##### Query dataset annotation #####

            logging.info("Annotating the query dataset...")
                    
            model_files = sorted(glob.glob(os.path.join(model_save_dir, "FFNN_model_Hydra_*.pt")))

            trainL = pd.read_csv(os.path.join(model_save_dir, "ct_train.csv"), index_col=0)

            known_cell_types = set(trainL['x'].values)

            if os.path.isfile(f"{split_folder}/adt_test.h5"):
                args.adt = f"{split_folder}/adt_test.h5"
            else:
                args.adt = "NULL" 
            if os.path.isfile(f"{split_folder}/atac_test.h5"):
                args.atac = f"{split_folder}/atac_test.h5"
            else:
                args.atac = "NULL"
            if os.path.isfile(f"{split_folder}/rna_test.h5"):
                args.rna = f"{split_folder}/rna_test.h5"
            else:
                args.rna = "NULL"

            (test_data, _, classify_dim) = process_data_annotation_query(args, feature_dir, split_folder)

            with open(os.path.join(split_folder, 'train_encoding.pkl'), 'rb') as f:
                train_encoding = pickle.load(f)
            
            transform_real_label, _ = real_label(args.cty, classify_dim)

            model_probabilities = []

            for model_file in model_files:
                model_load_path = model_file
                model = NN_Classifier(test_data.shape[1], len(known_cell_types))
                model.load_state_dict(torch.load(model_load_path, weights_only=False)) 
                model = model.to(device)
                model.eval()

                # Predict the labels for the test data
                with torch.no_grad():
                    output = model(test_data.to(device))
                    model_probabilities.append(output.cpu().numpy())  # For averaging, probability-based

            # Averaging the predicted probabilities from all models
            average_probabilities = np.mean(np.array(model_probabilities), axis=0)

            # Convert logits to probabilities
            average_probabilities = scipy.special.softmax(average_probabilities, axis=1)

            # Getting the class with the highest average probability
            test_pred = np.argmax(average_probabilities, axis=1)

            unique_cell_types = np.unique(transform_real_label)  

            unique_cell_types_dict = {}
            for cell_type in unique_cell_types:
                unique_cell_types_dict[cell_type] = train_encoding.get(cell_type, -1)
            
            # Reverse mapping from index to cell type name for predicted labels
            index_to_cell_type = {v: k for k, v in unique_cell_types_dict.items()}
            predicted_labels = [index_to_cell_type[idx] for idx in test_pred]

            # Creating a new DataFrame for the predicted labels similar to the ground truth format
            predicted_labels_df = pd.DataFrame(predicted_labels, columns=['x'])
            predicted_labels_df.index += 1  # Adjust index to start from 1 to match the example format
            predicted_labels_df.index.name = ""  # Match the format which has no name for index column

            # Saving annotation results
            cwd = os.getcwd()  # Get the current working directory
            Hydra_folder = os.path.join(cwd, f"Results/Annotation/Hydra-{args.num_classifiers}")
            if not os.path.exists(Hydra_folder):
                os.makedirs(Hydra_folder)

            # Save the predicted labels DataFrame as a CSV file
            predicted_labels_path = os.path.join(Hydra_folder, f"cell_type_predicted_Hydra-{args.num_classifiers}.csv")
            predicted_labels_df.to_csv(predicted_labels_path)

    return


##############################################

if __name__ == "__main__":
    # Logging to Standard Error
    Log_Format = "%(levelname)s - %(asctime)s - %(message)s \n"
    logging.basicConfig(stream = sys.stderr, format = Log_Format, level = logging.INFO)

    # Call the main function
    main()



#!/usr/bin/python

##############################################

# Manoj M Wagle (USydney, MIT CSAIL)

##############################################


import torch, os, sys
import torch.nn as nn
from tqdm import tqdm
from torch.autograd import Variable
import torch.nn.functional as F
from collections import Counter
import numpy as np
from .util import AverageMeter,accuracy,save_checkpoint,CrossEntropyLabelSmooth,KL_loss

device = torch.device("cuda" if torch.cuda.is_available()
                      else "mps" if getattr(torch, 'has_mps', False)
                      else "cpu")


##############################################

class CurriculumSampler(torch.utils.data.Sampler):
    def __init__(self, labels, num_epochs):
        self.labels = labels
        self.num_epochs = num_epochs
        self.current_epoch = 0

    def __iter__(self):
        weights = self._compute_weights()
        total = len(self.labels)
        indices = np.random.choice(total, size=total, p=weights)
        return iter(indices)

    def _compute_weights(self):
        label_counts = Counter(self.labels)
        max_count = max(label_counts.values())
        curriculum_ratio = self.current_epoch / self.num_epochs
        weights = [1 / (label_counts[label] * (1 + curriculum_ratio)) for label in self.labels]

        # Normalize weights
        weight_sum = sum(weights)
        weights = [weight / weight_sum for weight in weights]
        return weights
    
    def update_epoch(self, epoch):
        self.current_epoch = epoch



##############################################

def train_model(model, train_dl, lr, epochs, classify_dim=17, save_path="", save_filename="", feature_num=10000, use_balancing=False):
    criterion = nn.MSELoss().cuda()
    criterion_smooth_cty = CrossEntropyLabelSmooth().cuda()
    optimizer = torch.optim.Adam(model.parameters(), lr=lr)

    best_each_celltype_top1 = []
    best_each_celltype_num = []
    train_each_celltype_num = []
    
    for i in range(classify_dim):
        best_each_celltype_top1.append(0)
        best_each_celltype_num.append(0)
        train_each_celltype_num.append(0)

    best_loss = float('inf')
    no_improv = 0
    patience = 10

    # Curriculum learning sampler initialization
    if use_balancing:
        sampler = CurriculumSampler(train_dl.dataset.labels, epochs)
    else:
        sampler = None

    for epoch in tqdm(range(1, epochs + 1)):
        model.train()
        train_top1 = AverageMeter('Acc@1', ':6.2f')
        train_loss = 0.0
        train_batches = 0

        # Update the sampler for curriculum learning at the start of each epoch
        if use_balancing:
            sampler.update_epoch(epoch)

        for batch_sample in train_dl:
            optimizer.zero_grad()

            x = batch_sample['data'].to(device)
            x = Variable(x)
            x = torch.reshape(x, (x.size(0), -1))
            train_label = Variable(batch_sample['label']).to(device)

            encoder_out, x_cty, mu, var = model(x)
            
            # Calculating loss
            if isinstance(model, nn.DataParallel) and hasattr(model.module, 'classifiers'):
                x_cty = torch.mean(torch.stack([x_cty_i.to(device) for x_cty_i in x_cty]), dim=0)
            elif hasattr(model, 'classifiers'):
                x_cty = torch.mean(torch.stack([x_cty_i.to(device) for x_cty_i in x_cty]), dim=0)
            else:
                x_cty = x_cty.to(device)
                
            loss1 = criterion(encoder_out, x.to(device)) + 1 / feature_num * KL_loss(mu, var)
            loss2 = criterion_smooth_cty(x_cty, train_label.to(device))

            loss = 0.9 * loss1 + 0.1 * loss2
            loss.backward()
            optimizer.step()

            train_loss += loss.item()
            train_batches += 1
            
            train_pred1, = accuracy(x_cty, train_label, topk=(1,))
            train_top1.update(train_pred1[0], 1)

            if epoch == 1:
                for j in range(classify_dim):
                    if len(train_label[train_label == j]) != 0:
                        train_each_celltype_num[j] += len(train_label[train_label == j])

        average_train_loss = train_loss / train_batches
        
        # print(f"Epoch {epoch}/{epochs} - Train Loss: {average_train_loss:.4f}, Train Accuracy: {train_top1.avg:.4f}")
        
        if average_train_loss < best_loss:
            best_loss = average_train_loss
            no_improv = 0  # Reset the no improve counter
            model_to_save = model.module if isinstance(model, nn.DataParallel) else model
            torch.save({
                'epoch': epoch,
                'state_dict': model_to_save.state_dict(),
                'optimizer': optimizer.state_dict(),
            }, os.path.join(save_path, save_filename))
        else:
            no_improv += 1

        # If no improvement for 'patience' epochs, stop training
        if no_improv > patience:
            print("Early stopping at epoch:", epoch)
            break

    return model, best_loss, train_each_celltype_num



#!/usr/bin/python

##############################################

# Manoj M Wagle (USydney, MIT CSAIL)

##############################################


import torch
import torch.nn as nn
global mu
global var


##################################################

class Mish(nn.Module):
    def forward(self, x):
        return x * torch.tanh(torch.nn.functional.softplus(x))


##################################################

class LinBnDrop(nn.Sequential):
    """Module grouping `BatchNorm1d`, `Dropout` and `Linear` layers"""
    def __init__(self, n_in, n_out, bn=True, p=0., act=None, lin_first=True):
        self.p = p  # Store the dropout probability as an attribute
        layers = [nn.BatchNorm1d(n_out if lin_first else n_in)] if bn else []
        if p != 0: 
            layers.append(nn.Dropout(p))
        lin = [nn.Linear(n_in, n_out, bias=not bn)]
        if act is not None: lin.append(act)
        layers = lin+layers if lin_first else layers+lin
        super().__init__(*layers)


##################################################

class Encoder(nn.Module):
    """Encoder for 2 modalities (CITE-seq and SHARE-seq)"""
    def __init__(self, nfeatures_modality1=10703, nfeatures_modality2=192, hidden_modality1=185,  hidden_modality2=15, z_dim=128):
        super().__init__()
        self.nfeatures_modality1 = nfeatures_modality1
        self.nfeatures_modality2 = nfeatures_modality2
        self.encoder_modality1 = LinBnDrop(nfeatures_modality1, hidden_modality1, p=0.2, act=Mish())
        self.encoder_modality2 = LinBnDrop(nfeatures_modality2, hidden_modality2, p=0.2, act=Mish())
        self.encoder = LinBnDrop(hidden_modality1 + hidden_modality2, z_dim,  p=0.2, act=Mish())
        self.weights_modality1 = nn.Parameter(torch.rand((1,nfeatures_modality1)) * 0.001, requires_grad=True)
        self.weights_modality2 = nn.Parameter(torch.rand((1,nfeatures_modality2)) * 0.001, requires_grad=True)

        self.fc_mu = LinBnDrop(z_dim,z_dim, p=0.2)
        self.fc_var = LinBnDrop(z_dim,z_dim, p=0.2)

        self.mu = None
        self.var = None
        
    def reparameterize(self, mu, logvar):
        device = mu.device
        std = torch.exp(0.5 * logvar).to(device)
        eps = torch.randn_like(std).to(device)
        return eps * std + mu
    
    def forward(self, x):
        x_modality1 = self.encoder_modality1(x[:, :self.nfeatures_modality1]*self.weights_modality1)
        x_modality2 = self.encoder_modality2(x[:, self.nfeatures_modality1:]*self.weights_modality2)
        x = torch.cat([x_modality1, x_modality2], 1)
        x = self.encoder(x)
        self.mu = self.fc_mu(x)
        self.var = self.fc_var(x)
        x = self.reparameterize(self.mu, self.var)
        return x
    

class Decoder(nn.Module):
    """Decoder for 2 modalities (CITE-seq and SHARE-seq)"""
    def __init__(self, nfeatures_modality1=10703, nfeatures_modality2=192,  hidden_modality1=185,  hidden_modality2=15, z_dim=128):
        super().__init__()
        self.nfeatures_modality1 = nfeatures_modality1
        self.nfeatures_modality2 = nfeatures_modality2
        self.decoder1 = LinBnDrop(z_dim, nfeatures_modality1, act=Mish())
        self.decoder2 = LinBnDrop(z_dim, nfeatures_modality2,  act=Mish())
     
    def forward(self, x):
        x_rna = self.decoder1(x)
        x_adt = self.decoder2(x)
        x = torch.cat((x_rna,x_adt),1)
        return x


class Autoencoder_CITEseq_Step1(nn.Module):
    def __init__(self, nfeatures_rna=0, nfeatures_adt=0,  hidden_rna=185,  hidden_adt=15, z_dim=20,classify_dim=17, num_classifiers=1):
        """Autoencoder for CITE-seq Step-2"""
        super().__init__()
        self.encoder = Encoder(nfeatures_rna, nfeatures_adt, hidden_rna,  hidden_adt, z_dim)
        self.classifiers = nn.ModuleList([nn.Linear(z_dim, classify_dim) for _ in range(num_classifiers)])
        self.decoder = Decoder(nfeatures_rna, nfeatures_adt, hidden_rna,  hidden_adt, z_dim)
    
    def forward(self, x):
        x = self.encoder(x)
        x_cty = [classifier(x) for classifier in self.classifiers]
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty,mu,var


class Autoencoder_CITEseq_Step2(nn.Module):
    def __init__(self, nfeatures_rna=0, nfeatures_adt=0,  hidden_rna=185,  hidden_adt=15, z_dim=20,classify_dim=17):
        """Autoencoder for CITE-seq Step-1"""
        super().__init__()
        self.encoder = Encoder(nfeatures_rna, nfeatures_adt, hidden_rna,  hidden_adt, z_dim)
        self.classify = nn.Linear(z_dim, classify_dim)
        self.decoder = Decoder(nfeatures_rna, nfeatures_adt, hidden_rna,  hidden_adt, z_dim)
    
    def forward(self, x):
        x = self.encoder(x)
        x_cty = self.classify(x)
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty,mu,var


##################################################

class Autoencoder_SHAREseq_Step1(nn.Module):
    def __init__(self, nfeatures_rna=0, nfeatures_atac=0,  hidden_rna=185,  hidden_atac=15, z_dim=20,classify_dim=17, num_classifiers=1):
        """Autoencoder for SHARE-seq Step-1"""
        super().__init__()
        self.encoder = Encoder(nfeatures_rna, nfeatures_atac, hidden_rna,  hidden_atac, z_dim)
        self.classifiers = nn.ModuleList([nn.Linear(z_dim, classify_dim) for _ in range(num_classifiers)])
        self.decoder = Decoder(nfeatures_rna, nfeatures_atac, hidden_rna,  hidden_atac, z_dim)
    
    def forward(self, x):

        x = self.encoder(x)
        x_cty = [classifier(x) for classifier in self.classifiers]
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty,mu,var
    

class Autoencoder_SHAREseq_Step2(nn.Module):
    def __init__(self, nfeatures_rna=0, nfeatures_atac=0,  hidden_rna=185,  hidden_atac=15, z_dim=20,classify_dim=17):
        """Autoencoder for SHARE-seq Step-2"""
        super().__init__()
        self.encoder = Encoder(nfeatures_rna, nfeatures_atac, hidden_rna,  hidden_atac, z_dim)
        self.classify = nn.Linear(z_dim, classify_dim)
        self.decoder = Decoder(nfeatures_rna, nfeatures_atac, hidden_rna,  hidden_atac, z_dim)
    
    def forward(self, x):

        x = self.encoder(x)
        x_cty = self.classify(x)
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty,mu,var
    

##################################################

class Encoder_TEAseq(nn.Module):
    """Encoder for TEA-seq"""
    def __init__(self, nfeatures_rna=10703, nfeatures_adt=192, nfeatures_atac=192, hidden_rna=185, hidden_adt=30,  hidden_atac=185, z_dim=128):
        super().__init__()
        self.nfeatures_rna = nfeatures_rna
        self.nfeatures_adt = nfeatures_adt
        self.nfeatures_atac = nfeatures_atac

        self.encoder_rna = LinBnDrop(nfeatures_rna, hidden_rna, p=0.2, act=Mish())
        self.encoder_adt = LinBnDrop(nfeatures_adt, hidden_adt, p=0.2, act=Mish())
        self.encoder_atac = LinBnDrop(nfeatures_atac, hidden_atac, p=0.2, act=Mish())

        self.encoder = LinBnDrop(hidden_rna + hidden_adt +  hidden_atac, z_dim,  p=0.2, act=Mish())

        self.weights_rna = nn.Parameter(torch.rand((1,nfeatures_rna)) * 0.001, requires_grad=True)
        self.weights_adt = nn.Parameter(torch.rand((1,nfeatures_adt)) * 0.001, requires_grad=True)
        self.weights_atac = nn.Parameter(torch.rand((1,nfeatures_atac)) * 0.001, requires_grad=True)

        self.fc_mu = LinBnDrop(z_dim,z_dim)
        self.fc_var = LinBnDrop(z_dim,z_dim)
        self.mu = None
        self.var = None
        
    def reparameterize(self, mu, logvar):
        device = mu.device
        std = torch.exp(0.5 * logvar).to(device)
        eps = torch.randn_like(std).to(device)
        return eps * std + mu

    def forward(self, x):
        x_rna = self.encoder_rna(x[:, :self.nfeatures_rna]*self.weights_rna)
        x_adt = self.encoder_adt(x[:, self.nfeatures_rna:(self.nfeatures_rna+ self.nfeatures_adt)]*self.weights_adt)
        x_atac = self.encoder_atac(x[:, (self.nfeatures_rna+ self.nfeatures_adt):]*self.weights_atac)
        x = torch.cat([x_rna, x_adt, x_atac], 1)
        x = self.encoder(x)
        self.mu = self.fc_mu(x)
        self.var = self.fc_var(x)
        x = self.reparameterize(self.mu, self.var)
        return x
    

class Decoder_TEAseq(nn.Module):
    """Decoder for TEA-seq"""
    def __init__(self, nfeatures_rna=10703, nfeatures_adt=192, nfeatures_atac=10000, hidden_rna=185, hidden_adt=30, hidden_atac=185, z_dim=100):
        super().__init__()
        self.nfeatures_rna = nfeatures_rna
        self.nfeatures_adt = nfeatures_adt
        self.nfeatures_atac = nfeatures_atac
        self.decoder1 = LinBnDrop(z_dim, nfeatures_rna,  act=Mish())
        self.decoder2 = LinBnDrop(z_dim, nfeatures_adt,  act=Mish())
        self.decoder3 = LinBnDrop(z_dim, nfeatures_atac, act=Mish())

    def forward(self, x):
        x_rna = self.decoder1(x)
        x_adt = self.decoder2(x)
        x_atac = self.decoder3(x)
        x = torch.cat((x_rna,x_adt, x_atac), 1)
        return x
    

class Autoencoder_TEAseq_Step1(nn.Module):
    def __init__(self, nfeatures_rna=10000, nfeatures_adt=30, nfeatures_atac=10000, hidden_rna=185, hidden_adt=30,  hidden_atac=185, z_dim=100,classify_dim=17, num_classifiers=1):
        """ Autoencoder for TEA-seq Step-1"""
        super().__init__()
        self.encoder = Encoder_TEAseq(nfeatures_rna, nfeatures_adt, nfeatures_atac, hidden_rna,  hidden_adt, hidden_atac, z_dim)
        self.classifiers = nn.ModuleList([nn.Linear(z_dim, classify_dim) for _ in range(num_classifiers)])
        self.decoder = Decoder_TEAseq(nfeatures_rna, nfeatures_adt, nfeatures_atac, hidden_rna,  hidden_adt, hidden_atac, z_dim)

    def forward(self, x):
        x = self.encoder(x)
        x_cty = [classifier(x) for classifier in self.classifiers]
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty, mu, var
    

class Autoencoder_TEAseq_Step2(nn.Module):
    def __init__(self, nfeatures_rna=10000, nfeatures_adt=30, nfeatures_atac=10000, hidden_rna=185, hidden_adt=30, hidden_atac=185, z_dim=100, classify_dim=17):
        """ Autoencoder for TEA-seq Step-2"""
        super().__init__()
        self.encoder = Encoder_TEAseq(nfeatures_rna, nfeatures_adt, nfeatures_atac, hidden_rna, hidden_adt, hidden_atac, z_dim)
        self.classify = nn.Linear(z_dim, classify_dim)
        self.decoder = Decoder_TEAseq(nfeatures_rna, nfeatures_adt, nfeatures_atac, hidden_rna, hidden_adt, hidden_atac, z_dim)

    def forward(self, x):
        x = self.encoder(x)
        x_cty = self.classify(x)
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty, mu, var


##################################################
        
class Encoder_RNA_seq(nn.Module):
    """Encoder for RNA-seq"""
    def __init__(self, nfeatures_modality1=10703, hidden_modality1=185,  z_dim=128):
        super().__init__()
        self.nfeatures_modality1 = nfeatures_modality1
        self.encoder_modality1 = LinBnDrop(nfeatures_modality1, hidden_modality1, p=0.2, act=Mish())
        self.encoder = LinBnDrop(hidden_modality1, z_dim,  p=0.2, act=Mish())
        self.weights_modality1 = nn.Parameter(torch.rand((1,nfeatures_modality1)) * 0.001, requires_grad=True)

        self.fc_mu = LinBnDrop(z_dim,z_dim, p=0.2)
        self.fc_var = LinBnDrop(z_dim,z_dim, p=0.2)

        self.mu = None
        self.var = None
        
    def reparameterize(self, mu, logvar):
        device = mu.device
        std = torch.exp(0.5 * logvar).to(device)
        eps = torch.randn_like(std).to(device)
        return eps * std + mu
    
    def forward(self, x):
        x = self.encoder_modality1(x*self.weights_modality1)
        x = self.encoder(x)
        self.mu = self.fc_mu(x)
        self.var = self.fc_var(x).to(self.mu.device)
        x = self.reparameterize(self.mu, self.var)
        return x
    

class Decoder_RNA_seq(nn.Module):
    """Decoder for RNA-seq """
    def __init__(self, nfeatures_modality1=10703, hidden_modality1=185, z_dim=128):
        super().__init__()
        self.nfeatures_modality1 = nfeatures_modality1
        self.decoder1 = LinBnDrop(z_dim, nfeatures_modality1, act=Mish())

    def forward(self, x):
        x = self.decoder1(x)
        return x

    
class Autoencoder_RNAseq_Step1(nn.Module):
    def __init__(self, nfeatures_rna=0, hidden_rna=185, z_dim=20,classify_dim=17, num_classifiers=1):
        """ Autoencoder for RNA-seq Step-1"""
        super().__init__()
        self.encoder = Encoder_RNA_seq(nfeatures_rna, hidden_rna, z_dim)
        self.classifiers = nn.ModuleList([nn.Linear(z_dim, classify_dim) for _ in range(num_classifiers)])
        self.decoder = Decoder_RNA_seq(nfeatures_rna, hidden_rna, z_dim)
        
    def forward(self, x):
        x = self.encoder(x)
        device = x.device
        x_cty = [classifier(x.to(device)) for classifier in self.classifiers]
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty, mu, var
    

class Autoencoder_RNAseq_Step2(nn.Module):
    def __init__(self, nfeatures_rna=0, hidden_rna=185, z_dim=20,classify_dim=17):
        """ Autoencoder for RNA-seq Step-2"""
        super().__init__()
        self.encoder = Encoder_RNA_seq(nfeatures_rna, hidden_rna, z_dim)
        self.classify = nn.Linear(z_dim, classify_dim)
        self.decoder = Decoder_RNA_seq(nfeatures_rna, hidden_rna, z_dim)
        
    def forward(self, x):
        x = self.encoder(x)
        x_cty = self.classify(x)
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty, mu, var
    

##################################################
        
class Encoder_ADT_seq(nn.Module):
    """Encoder for ADT-seq"""
    def __init__(self, nfeatures_modality1=192, hidden_modality1=30,  z_dim=128):
        super().__init__()
        self.nfeatures_modality1 = nfeatures_modality1
        self.encoder_modality1 = LinBnDrop(nfeatures_modality1, hidden_modality1, p=0.2, act=Mish())
        self.encoder = LinBnDrop(hidden_modality1, z_dim,  p=0.2, act=Mish())
        self.weights_modality1 = nn.Parameter(torch.rand((1,nfeatures_modality1)) * 0.001, requires_grad=True)

        self.fc_mu = LinBnDrop(z_dim,z_dim, p=0.2)
        self.fc_var = LinBnDrop(z_dim,z_dim, p=0.2)

        self.mu = None
        self.var = None
        
    def reparameterize(self, mu, logvar):
        device = mu.device
        std = torch.exp(0.5 * logvar).to(device)
        eps = torch.randn_like(std).to(device)
        return eps * std + mu
    
    def forward(self, x):
        x = self.encoder_modality1(x*self.weights_modality1)
        x = self.encoder(x)
        self.mu = self.fc_mu(x)
        self.var = self.fc_var(x).to(self.mu.device)
        x = self.reparameterize(self.mu, self.var)
        return x
    

class Decoder_ADT_seq(nn.Module):
    """Decoder for ADT-seq """
    def __init__(self, nfeatures_modality1=192, hidden_modality1=30, z_dim=128):
        super().__init__()
        self.nfeatures_modality1 = nfeatures_modality1
        self.decoder1 = LinBnDrop(z_dim, nfeatures_modality1, act=Mish())

    def forward(self, x):
        x = self.decoder1(x)
        return x

    
class Autoencoder_ADTseq_Step1(nn.Module):
    def __init__(self, nfeatures_adt=0, hidden_adt=30, z_dim=20, classify_dim=17, num_classifiers=1):
        """ Autoencoder for ADT-seq Step-1"""
        super().__init__()
        self.encoder = Encoder_RNA_seq(nfeatures_adt, hidden_adt, z_dim)
        self.classifiers = nn.ModuleList([nn.Linear(z_dim, classify_dim) for _ in range(num_classifiers)])
        self.decoder = Decoder_RNA_seq(nfeatures_adt, hidden_adt, z_dim)
        
    def forward(self, x):
        x = self.encoder(x)
        device = x.device
        x_cty = [classifier(x.to(device)) for classifier in self.classifiers]
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty, mu, var
    

class Autoencoder_ADTseq_Step2(nn.Module):
    def __init__(self, nfeatures_adt=0, hidden_adt=30, z_dim=20,classify_dim=17):
        """ Autoencoder for ADT-seq Step-2"""
        super().__init__()
        self.encoder = Encoder_RNA_seq(nfeatures_adt, hidden_adt, z_dim)
        self.classify = nn.Linear(z_dim, classify_dim)
        self.decoder = Decoder_RNA_seq(nfeatures_adt, hidden_adt, z_dim)
        
    def forward(self, x):
        x = self.encoder(x)
        x_cty = self.classify(x)
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty, mu, var
    

##################################################
        
class Encoder_ATAC_seq(nn.Module):
    """Encoder for ATAC-seq"""
    def __init__(self, nfeatures_modality1=192, hidden_modality1=185,  z_dim=128):
        super().__init__()
        self.nfeatures_modality1 = nfeatures_modality1
        self.encoder_modality1 = LinBnDrop(nfeatures_modality1, hidden_modality1, p=0.2, act=Mish())
        self.encoder = LinBnDrop(hidden_modality1, z_dim,  p=0.2, act=Mish())
        self.weights_modality1 = nn.Parameter(torch.rand((1,nfeatures_modality1)) * 0.001, requires_grad=True)

        self.fc_mu = LinBnDrop(z_dim,z_dim, p=0.2)
        self.fc_var = LinBnDrop(z_dim,z_dim, p=0.2)

        self.mu = None
        self.var = None
        
    def reparameterize(self, mu, logvar):
        device = mu.device
        std = torch.exp(0.5 * logvar).to(device)
        eps = torch.randn_like(std).to(device)
        return eps * std + mu
    
    def forward(self, x):
        x = self.encoder_modality1(x*self.weights_modality1)
        x = self.encoder(x)
        self.mu = self.fc_mu(x)
        self.var = self.fc_var(x).to(self.mu.device)
        x = self.reparameterize(self.mu, self.var)
        return x
    

class Decoder_ATAC_seq(nn.Module):
    """Decoder for ATAC-seq """
    def __init__(self, nfeatures_modality1=192, hidden_modality1=185, z_dim=128):
        super().__init__()
        self.nfeatures_modality1 = nfeatures_modality1
        self.decoder1 = LinBnDrop(z_dim, nfeatures_modality1, act=Mish())

    def forward(self, x):
        x = self.decoder1(x)
        return x

    
class Autoencoder_ATACseq_Step1(nn.Module):
    def __init__(self, nfeatures_atac=0, hidden_atac=185, z_dim=20, classify_dim=17, num_classifiers=1):
        """ Autoencoder for ATAC-seq Step-1"""
        super().__init__()
        self.encoder = Encoder_RNA_seq(nfeatures_atac, hidden_atac, z_dim)
        self.classifiers = nn.ModuleList([nn.Linear(z_dim, classify_dim) for _ in range(num_classifiers)])
        self.decoder = Decoder_RNA_seq(nfeatures_atac, hidden_atac, z_dim)
        
    def forward(self, x):
        x = self.encoder(x)
        device = x.device
        x_cty = [classifier(x.to(device)) for classifier in self.classifiers]
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty, mu, var
    

class Autoencoder_ATACseq_Step2(nn.Module):
    def __init__(self, nfeatures_atac=0, hidden_atac=185, z_dim=20,classify_dim=17):
        """ Autoencoder for ATAC-seq Step-2"""
        super().__init__()
        self.encoder = Encoder_RNA_seq(nfeatures_atac, hidden_atac, z_dim)
        self.classify = nn.Linear(z_dim, classify_dim)
        self.decoder = Decoder_RNA_seq(nfeatures_atac, hidden_atac, z_dim)
        
    def forward(self, x):
        x = self.encoder(x)
        x_cty = self.classify(x)
        x = self.decoder(x)
        mu = self.encoder.mu
        var = self.encoder.var
        return x, x_cty, mu, var



#!/usr/bin/python

##############################################

# Manoj M Wagle (USydney, MIT CSAIL)

##############################################

import subprocess
import sys
from setuptools import setup, find_namespace_packages

setup(
    name="hydra-tools",
    version="0.1.0",
    packages=find_namespace_packages(),
    install_requires=[
        'anndata',
        'captum',
        'h5py',
        'libpng-bins',
        'torch',
        'torchvision',
        'torchaudio',
        'pandas',
        'scipy',
        'scikit-learn',
        'scanpy',
        'tqdm',
        'numpy',
        'numba'
    ],
    entry_points={
        'console_scripts': [
            'hydra=hydra.Hydra:main',
        ],
    },
    include_package_data=True,
    description='Thank you for using Hydra 😄, an interpretable deep generative tool for single-cell omics. Please refer to the full documentation available at https://sydneybiox.github.io/Hydra/ for detailed usage instructions. If you encounter any issues running the tool - Please open an issue on Github, and we will get back to you as soon as possible!!',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/SydneyBioX/Hydra',
    author='Manoj M Wagle',
    author_email='mwag8019@uni.sydney.edu.au',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8'
    ],
)



#!/usr/bin/python

##############################################

# Manoj M Wagle (USydney, MIT CSAIL)

##############################################


import pandas as pd
import numpy as np
import torch
import torch.nn as nn
from torch.utils.data import Dataset, DataLoader
import torch.nn.functional as F
import scanpy as sc
import random, os, re
from torch.autograd import Variable
import h5py
import scipy
import pickle

device = torch.device("cuda" if torch.cuda.is_available()
                      else "mps" if getattr(torch, 'has_mps', False)
                      else "cpu")

FloatTensor = torch.FloatTensor
LongTensor = torch.LongTensor


##################################################

def setup_seed(seed):
     torch.manual_seed(seed)
     torch.cuda.manual_seed_all(seed)
     np.random.seed(seed)
     random.seed(seed)
     torch.backends.cudnn.deterministic = True
     torch.backends.cudnn.benchmark = False
     os.environ['PYTHONHASHSEED']=str(seed)


##################################################

def real_label(label_path, classify_dim):
    output_v = []
    output_dict = {}
    label = pd.read_csv(label_path, header=None, index_col=False)
    label_real = label.iloc[1:(label.shape[0]), 1]
    (label_num, _) = read_fs_label(label_path)
    
    # Get unique labels in sorted order
    unique_labels = sorted(np.unique(label_num.cpu()))

    for i in unique_labels:    
        idx = np.array(torch.where(label_num==i)[0].cpu()).astype('int32')
        if len(idx) > 0:
            temp = label_real[idx[0] + 1]
            output_v.append(temp)
            count = sum(label_real == temp)  # Count the occurrences of temp in label_real
            output_dict[i] = (temp, count)
    return output_v, output_dict


##################################################

def Index2Label(label_path,classify_dim):
    label = pd.read_csv(label_path,header=None,index_col=False)  
    label_real = label.iloc[1:(label.shape[0]),1]
    (label_num, _) = read_fs_label(label_path)
    index_to_label = {}
    for i in range(classify_dim):    
        temp = label_real[np.array(torch.where(label_num==i)[0][0].cpu()).astype('int32')+1]
        index_to_label[i] = temp
    return index_to_label


##################################################

def read_fs_label(label_path):
    label_fs = pd.read_csv(label_path, header=None, index_col=False)
    label_fs_codes = pd.Categorical(label_fs.iloc[1:(label_fs.shape[0]),1]).codes
    label_fs_names = label_fs.iloc[1:, 1].tolist()
    # for x,y in zip(label_fs_names[:15], label_fs_codes[:15]):
    #     print(x, " : ", y)
    # sys.exit()
    label_fs_codes = np.array(label_fs_codes[:]).astype('int32')
    label_fs_codes = torch.from_numpy(label_fs_codes)
    label_fs_codes = label_fs_codes.type(LongTensor)
    return label_fs_codes, label_fs_names


##################################################

def read_fs_label_test(label_path, encoding = None):
    label_fs = pd.read_csv(label_path, header=None, index_col=False, skiprows=1)
    label_fs_names = label_fs[1].tolist()
    label_fs['encoded'] = label_fs[1].map(encoding)
    label_fs_codes = torch.from_numpy(label_fs['encoded'].values).type(LongTensor)
    # for x,y in zip(label_fs_names[:15], label_fs_codes[:15]):
    #     print(x, " : ", y)
    # sys.exit()
    return label_fs_codes, label_fs_names


##################################################

class MyDataset(Dataset):
    def __init__(self, data, label):
        self.data = data
        self.label = label
        self.labels = label

    def __getitem__(self, index): 
        img, target = self.data[index,:], self.label[index]
        sample = {'data': img, 'label': target}
        return sample

    def __len__(self):
        return len(self.data)
        

##################################################

class ToTensor(object):
    def __call__(self, sample):
        data,label = sample['data'], sample['label']
        data = data.transpose((1, 0))
        return {'data': torch.from_numpy(data),
                'label': torch.from_numpy(label),
               }
    

##################################################
               
class AverageMeter(object):
    """Computes and stores the average and current value"""
    def __init__(self, name, fmt=':f'):
        self.name = name
        self.fmt = fmt
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count

    def __str__(self):
        fmtstr = '{name} {val' + self.fmt + '} ({avg' + self.fmt + '})'
        return fmtstr.format(**self.__dict__)


##################################################

def save_checkpoint(state, filename):
    save_dir = os.path.dirname(filename)
    if not os.path.exists(save_dir):
        os.makedirs(save_dir)
    torch.save(state, filename)
    

##################################################

class CrossEntropyLabelSmooth(nn.Module):

  def __init__(self, num_classes=17, epsilon=0.1):
    super(CrossEntropyLabelSmooth, self).__init__()
    self.num_classes = num_classes
    self.epsilon = epsilon
    self.logsoftmax = nn.LogSoftmax(dim=1)

  def forward(self, inputs, targets):
    log_probs = self.logsoftmax(inputs)
    targets = torch.zeros_like(log_probs).scatter_(1, targets.unsqueeze(1), 1)
    targets = (1 - self.epsilon) * targets + self.epsilon / self.num_classes
    loss = (-targets * log_probs).mean(0).sum()
    return loss


##################################################

def read_h5_data(data_path):
    data = h5py.File(data_path,"r")
    h5_data = data['matrix/data']
    sparse_data = scipy.sparse.csr_matrix(np.array(h5_data).transpose())
    data_fs = torch.from_numpy(np.array(sparse_data.todense()))
    data_fs = Variable(data_fs.type(FloatTensor))
    return data_fs


##################################################

def get_encodings(model, dl):
    model.eval()
    encodings = []
    ori_data = []
    label = [] 
    with torch.no_grad():
        for i, batch_sample in enumerate(dl):
                x = batch_sample['data']
                x = Variable(x.type(FloatTensor))
                x = torch.reshape(x,(x.size(0),-1))
                rna_valid_label = batch_sample['label']
                rna_valid_label = Variable(rna_valid_label.type(LongTensor))
            
                x_prime = model.encoder(x.to(device))
                encodings.append(x_prime)
                ori_data.append(x)
                label.append(rna_valid_label)
                
                
    return torch.cat(encodings, dim=0),torch.cat(ori_data,dim=0),torch.cat(label,dim=0)


##################################################

def get_decodings(model, dl):
    model.eval()
    decodings = []
    ori_data = []
    with torch.no_grad():
        for i, batch_sample in enumerate(dl):
                x = batch_sample['data']
                x = Variable(x.type(FloatTensor))
                x = torch.reshape(x,(x.size(0),-1))
                rna_valid_label = batch_sample['label']
                rna_valid_label = Variable(rna_valid_label.type(LongTensor))
                x_prime, x_cty,mu,var = model(x.to(device))
                decodings.append(x_prime)
                ori_data.append(x)                
    return torch.cat(decodings, dim=0),torch.cat(ori_data,dim=0)


##################################################

def get_simulated_data_random_generation(model, dl):
    model.eval()
    decodings = []
    ori_data = []
    label = []
    for i, batch_sample in enumerate(dl):
            x = batch_sample['data']
            x = Variable(x.type(FloatTensor))
            x = torch.reshape(x,(x.size(0),-1))
            rna_valid_label = batch_sample['label']
            rna_valid_label = Variable(rna_valid_label.type(LongTensor))
            
            x_prime, x_cty,mu,var = model(x.to(device))
            decodings.append(x_prime)
            ori_data.append(x)
            label.append(rna_valid_label)
                
    return torch.cat(decodings, dim=0),torch.cat(label,dim=0)


##################################################

def get_simulated_data(model, dl):
    model.eval()
    decodings = []
    ori_data = []
    label = []
    with torch.no_grad():
        for i, batch_sample in enumerate(dl):
                x = batch_sample['data']
                x = Variable(x.type(FloatTensor))
                x = torch.reshape(x,(x.size(0),-1))
                rna_valid_label = batch_sample['label']
                rna_valid_label = Variable(rna_valid_label.type(LongTensor))
            
                x_prime, x_cty,mu,var = model(x.to(device))
                decodings.append(x_prime)
                ori_data.append(x)
                label.append(rna_valid_label)
                
    return torch.cat(decodings, dim=0),torch.cat(label,dim=0),torch.cat(ori_data,dim=0)
    

##################################################

def KL_loss(mu, logvar):
    KLD = -0.5 * torch.mean(1 + logvar - mu**2 -  logvar.exp())
    return  KLD


##################################################

def get_vae_simulated_data_from_sampling(model, dl):
    model.eval()
    latent = []
    ori_data = []
    label = []
    decodings = []
    with torch.no_grad():
        for i, batch_sample in enumerate(dl):
                x = batch_sample['data']
                x = Variable(x.type(FloatTensor))
                x = torch.reshape(x,(x.size(0),-1))
                
                ori_data.append(x)
                rna_valid_label = batch_sample['label']
                rna_valid_label = Variable(rna_valid_label.type(LongTensor))
                y, x_cty,mu,var = model(x.to(device))
                decodings.append(y)
                label.append(rna_valid_label)
                
    return torch.cat(decodings, dim=0),torch.cat(label,dim=0),torch.cat(ori_data,dim=0)


##############################################

def load_and_preprocess_data(args, setting=""):
    if args.adt != "NULL" and args.atac != "NULL":

        mode = "scRNA+scADT+scATAC"

        if setting == 'train':
            train_rna_data_path = args.rna
            train_adt_data_path = args.adt
            train_atac_data_path = args.atac
            train_label_path = args.cty

            train_rna_data = read_h5_data(train_rna_data_path)
            train_adt_data = read_h5_data(train_adt_data_path)
            train_atac_data = read_h5_data(train_atac_data_path)
            (train_label, _) = read_fs_label(train_label_path)
            classify_dim = (max(train_label)+1).cpu().numpy()

            _, label_to_name_mapping = real_label(train_label_path, classify_dim)
            
            # Get the number of features
            nfeatures_rna = train_rna_data.shape[1]
            nfeatures_adt = train_adt_data.shape[1]
            nfeatures_atac = train_atac_data.shape[1]
            
            # Total features count 
            feature_num = nfeatures_rna + nfeatures_adt + nfeatures_atac

            # Prepare training data
            train_data = torch.cat((train_rna_data, train_adt_data, train_atac_data), 1) 
            train_transformed_dataset = MyDataset(train_data, train_label)
            train_dl = DataLoader(train_transformed_dataset, batch_size = args.batch_size, shuffle = True, num_workers = 0, drop_last = False)
        else:
            test_rna_data_path = args.rna
            test_adt_data_path = args.adt
            test_atac_data_path = args.atac
            test_label_path = args.cty

            test_rna_data = read_h5_data(test_rna_data_path)
            test_adt_data = read_h5_data(test_adt_data_path)
            test_atac_data = read_h5_data(test_atac_data_path)
            (test_label, _) = read_fs_label(test_label_path)
            classify_dim = (max(test_label)+1).cpu().numpy()

            _, label_to_name_mapping = real_label(test_label_path, classify_dim)
            
            # Get the number of features
            nfeatures_rna = test_rna_data.shape[1]
            nfeatures_adt = test_adt_data.shape[1]
            nfeatures_atac = test_atac_data.shape[1]

            # Total features count 
            feature_num = nfeatures_rna + nfeatures_adt + nfeatures_atac

            # Prepare test data
            test_data = torch.cat((test_rna_data, test_adt_data, test_atac_data), 1) 
            test_transformed_dataset = MyDataset(test_data, test_label)
            test_dl = DataLoader(test_transformed_dataset, batch_size = args.batch_size, shuffle = False, num_workers = 0, drop_last = False)


    if args.adt != "NULL" and args.atac == "NULL":

        mode = "scRNA+scADT"

        if setting == 'train':
            train_rna_data_path = args.rna
            train_adt_data_path = args.adt
            train_label_path = args.cty

            train_rna_data = read_h5_data(train_rna_data_path)
            train_adt_data = read_h5_data(train_adt_data_path)
            (train_label, _) = read_fs_label(train_label_path)
            classify_dim = (max(train_label)+1).cpu().numpy()

            _, label_to_name_mapping = real_label(train_label_path, classify_dim)
            
            # Get the number of features
            nfeatures_rna = train_rna_data.shape[1]
            nfeatures_adt = train_adt_data.shape[1]
            
            # Total features count 
            feature_num = nfeatures_rna + nfeatures_adt

            # Prepare training data
            train_data = torch.cat((train_rna_data, train_adt_data), 1) 
            train_transformed_dataset = MyDataset(train_data, train_label)
            train_dl = DataLoader(train_transformed_dataset, batch_size = args.batch_size, shuffle = True, num_workers = 0, drop_last = False)
        else:
            test_rna_data_path = args.rna
            test_adt_data_path = args.adt
            test_label_path = args.cty

            test_rna_data = read_h5_data(test_rna_data_path)
            test_adt_data = read_h5_data(test_adt_data_path)
            (test_label, _) = read_fs_label(test_label_path)
            classify_dim = (max(test_label)+1).cpu().numpy()

            _, label_to_name_mapping = real_label(test_label_path, classify_dim)
            
            # Get the number of features
            nfeatures_rna = test_rna_data.shape[1]
            nfeatures_adt = test_adt_data.shape[1]

            # Total features count 
            feature_num = nfeatures_rna + nfeatures_adt

            # Prepare test data
            test_data = torch.cat((test_rna_data, test_adt_data), 1) 
            test_transformed_dataset = MyDataset(test_data, test_label)
            test_dl = DataLoader(test_transformed_dataset, batch_size = args.batch_size, shuffle = False, num_workers = 0, drop_last = False)


    if args.adt == "NULL" and args.atac != "NULL":

        mode = "scRNA+scATAC"

        if setting == 'train':
            train_rna_data_path = args.rna
            train_atac_data_path = args.atac
            train_label_path = args.cty

            train_rna_data = read_h5_data(train_rna_data_path)
            train_atac_data = read_h5_data(train_atac_data_path)
            (train_label, _) = read_fs_label(train_label_path)
            classify_dim = (max(train_label)+1).cpu().numpy()

            _, label_to_name_mapping = real_label(train_label_path, classify_dim)
            
            # Get the number of features
            nfeatures_rna = train_rna_data.shape[1]
            nfeatures_atac = train_atac_data.shape[1]

            # Total features count 
            feature_num = nfeatures_rna + nfeatures_atac

            # Prepare training data
            train_data = torch.cat((train_rna_data, train_atac_data), 1) 
            train_transformed_dataset = MyDataset(train_data, train_label)
            train_dl = DataLoader(train_transformed_dataset, batch_size = args.batch_size, shuffle = True, num_workers = 0, drop_last = False)
        else:
            test_rna_data_path = args.rna
            test_atac_data_path = args.atac
            test_label_path = args.cty

            test_rna_data = read_h5_data(test_rna_data_path)
            test_atac_data = read_h5_data(test_atac_data_path)
            (test_label, _) = read_fs_label(test_label_path)
            classify_dim = (max(test_label)+1).cpu().numpy()

            _, label_to_name_mapping = real_label(test_label_path, classify_dim)
            
            # Get the number of features
            nfeatures_rna = test_rna_data.shape[1]
            nfeatures_atac = test_atac_data.shape[1]

            # Total features count 
            feature_num = nfeatures_rna + nfeatures_atac

            # Prepare test data
            test_data = torch.cat((test_rna_data, test_atac_data), 1) 
            test_transformed_dataset = MyDataset(test_data, test_label)
            test_dl = DataLoader(test_transformed_dataset, batch_size = args.batch_size, shuffle = False, num_workers = 0, drop_last = False)


    if args.adt == "NULL" and args.atac == "NULL":

        mode = "scRNA-seq"

        if setting == 'train':
            train_rna_data_path = args.rna
            train_label_path = args.cty

            train_rna_data = read_h5_data(train_rna_data_path)
            (train_label, _) = read_fs_label(train_label_path)
            classify_dim = (max(train_label)+1).cpu().numpy()

            _, label_to_name_mapping = real_label(train_label_path, classify_dim)
            
            # Get the number of features
            nfeatures_rna = train_rna_data.shape[1]

            # Total features count 
            feature_num = nfeatures_rna

            # Prepare training data
            train_data = train_rna_data
            train_transformed_dataset = MyDataset(train_data, train_label)
            train_dl = DataLoader(train_transformed_dataset, batch_size = args.batch_size, shuffle = True, num_workers = 0, drop_last = False)
        else:
            test_rna_data_path = args.rna
            test_label_path = args.cty

            test_rna_data = read_h5_data(test_rna_data_path)
            (test_label, _) = read_fs_label(test_label_path)
            classify_dim = (max(test_label)+1).cpu().numpy()

            _, label_to_name_mapping = real_label(test_label_path, classify_dim)
            
            # Get the number of features
            nfeatures_rna = test_rna_data.shape[1]
            
            # Total features count 
            feature_num = nfeatures_rna

            # Prepare test data
            test_data = test_rna_data
            test_transformed_dataset = MyDataset(test_data, test_label)
            test_dl = DataLoader(test_transformed_dataset, batch_size = args.batch_size, shuffle = False, num_workers = 0, drop_last = False)
    
    if args.rna == "NULL" and args.atac == "NULL":

        mode = "scADT-seq"

        if setting == 'train':
            train_adt_data_path = args.adt
            train_label_path = args.cty
            train_adt_data = read_h5_data(train_adt_data_path)
            (train_label, _) = read_fs_label(train_label_path)
            classify_dim = (max(train_label)+1).cpu().numpy()
            _, label_to_name_mapping = real_label(train_label_path, classify_dim)

            # Get the number of features
            nfeatures_adt = train_adt_data.shape[1]
            
            # Total features count 
            feature_num = nfeatures_adt
            
            # Prepare training data
            train_data = train_adt_data
            train_transformed_dataset = MyDataset(train_data, train_label)
            train_dl = DataLoader(train_transformed_dataset, batch_size = args.batch_size, shuffle = True, num_workers = 0, drop_last = False)
        else:
            test_adt_data_path = args.adt
            test_label_path = args.cty
            test_adt_data = read_h5_data(test_adt_data_path)
            (test_label, _) = read_fs_label(test_label_path)
            classify_dim = (max(test_label)+1).cpu().numpy()
            _, label_to_name_mapping = real_label(test_label_path, classify_dim)

            # Get the number of features
            nfeatures_adt = test_adt_data.shape[1]

            # Total features count 
            feature_num = nfeatures_adt
            
            # Prepare test data
            test_data = test_adt_data
            test_transformed_dataset = MyDataset(test_data, test_label)
            test_dl = DataLoader(test_transformed_dataset, batch_size = args.batch_size, shuffle = False, num_workers = 0, drop_last = False)
    

    if args.rna == "NULL" and args.adt == "NULL":

        mode = "scATAC-seq"
    
        if setting == 'train':
            train_atac_data_path = args.atac
            train_label_path = args.cty
            train_atac_data = read_h5_data(train_atac_data_path)
            (train_label, _) = read_fs_label(train_label_path)
            classify_dim = (max(train_label)+1).cpu().numpy()
            _, label_to_name_mapping = real_label(train_label_path, classify_dim)

            # Get the number of features
            nfeatures_atac = train_atac_data.shape[1]
            
            # Total features count 
            feature_num = nfeatures_atac
            
            # Prepare training data
            train_data = train_atac_data
            train_transformed_dataset = MyDataset(train_data, train_label)
            train_dl = DataLoader(train_transformed_dataset, batch_size = args.batch_size, shuffle = True, num_workers = 0, drop_last = False)
        else:
            test_atac_data_path = args.atac
            test_label_path = args.cty
            test_atac_data = read_h5_data(test_atac_data_path)
            (test_label, _) = read_fs_label(test_label_path)
            classify_dim = (max(test_label)+1).cpu().numpy()
            _, label_to_name_mapping = real_label(test_label_path, classify_dim)

            # Get the number of features
            nfeatures_atac = test_atac_data.shape[1]

            # Total features count 
            feature_num = nfeatures_atac
            
            # Prepare test data
            test_data = test_atac_data
            test_transformed_dataset = MyDataset(test_data, test_label)
            test_dl = DataLoader(test_transformed_dataset, batch_size = args.batch_size, shuffle = False, num_workers = 0, drop_last = False)


    if mode == "scRNA+scADT+scATAC":
        if setting == 'train':
            return train_data, train_dl, train_label, mode, classify_dim, nfeatures_rna, nfeatures_adt, nfeatures_atac, feature_num, label_to_name_mapping
        else:
            return test_data, test_dl, test_label, mode, classify_dim, nfeatures_rna, nfeatures_adt, nfeatures_atac, feature_num, label_to_name_mapping

    if mode == "scRNA+scADT":
        if setting == 'train':
            return train_data, train_dl, train_label, mode, classify_dim, nfeatures_rna, nfeatures_adt, feature_num, label_to_name_mapping
        else:
            return test_data, test_dl, test_label, mode, classify_dim, nfeatures_rna, nfeatures_adt, feature_num, label_to_name_mapping

    if mode == "scRNA+scATAC":
        if setting == 'train':
            return train_data, train_dl, train_label, mode, classify_dim, nfeatures_rna, nfeatures_atac, feature_num, label_to_name_mapping
        else:
            return test_data, test_dl, test_label, mode, classify_dim, nfeatures_rna, nfeatures_atac, feature_num, label_to_name_mapping

    if mode == "scRNA-seq":
        if setting == 'train':
            return train_data, train_dl, train_label, mode, classify_dim, nfeatures_rna, feature_num, label_to_name_mapping
        else:
            return test_data, test_dl, test_label, mode, classify_dim, nfeatures_rna, feature_num, label_to_name_mapping

    if mode == "scADT-seq":
        if setting == 'train':
            return train_data, train_dl, train_label, mode, classify_dim, nfeatures_adt, feature_num, label_to_name_mapping
        else:
            return test_data, test_dl, test_label, mode, classify_dim, nfeatures_adt, feature_num, label_to_name_mapping

    if mode == "scATAC-seq":
        if setting == 'train':
            return train_data, train_dl, train_label, mode, classify_dim, nfeatures_atac, feature_num, label_to_name_mapping
        else:
            return test_data, test_dl, test_label, mode, classify_dim, nfeatures_atac, feature_num, label_to_name_mapping


##################################################

def process_data_annotation_train(args, feature_dir, split_folder):

    label_path = args.cty

    # Load data
    train_data = torch.load(args.balanced_data)
    train_label = torch.load(args.balanced_label)

    (label_fs_codes, label_fs_names) = read_fs_label(label_path)
    # train_encoding dictionary
    train_encoding = {}
    for name, code in zip(label_fs_names, label_fs_codes):
        if name not in train_encoding:
            train_encoding[name] = code.cpu().numpy().item()

    # Add the 'Unknown' key to train_encoding with the next available integer code
    train_encoding['Unknown'] = max(train_encoding.values()) + 1
    with open(os.path.join(split_folder, 'train_encoding.pkl'), 'wb') as f:
        pickle.dump(train_encoding, f)

    classify_dim = (max(train_label)+1).cpu().numpy()

    ### Hybrid Feature Selection Strategy
    # Load the original labels
    df_original = pd.read_csv(label_path)
    cell_type_counts = df_original['x'].value_counts()

    # Determine the threshold for rare and major classes
    median_count = cell_type_counts.median()
    rare_cell_types = cell_type_counts[cell_type_counts < median_count].index.tolist()

    top_features = []

    # Process each feature file
    for filename in os.listdir(feature_dir):
        if filename.endswith('.csv'):
            filepath = os.path.join(feature_dir, filename)
            
            # Read the CSV file
            df = pd.read_csv(filepath)
            
            # Get cell type from the filename
            cell_type = re.search('fs\.(.*?)_', filename).group(1)
            
            # number of top features to extract
            if cell_type in rare_cell_types:
                num_features = 100  # number of top features for minor cell types
            else:
                num_features = 100  # number of top features for major cell types
            
            # Get the top features
            top_features_file = df['Feature Name'].head(num_features).tolist()
            
            # Add these features to the list
            top_features.extend(top_features_file)

    # Get the unique set of top features across all files
    top_features = list(set(top_features))

    # Extract indices from the features and sort them
    top_feature_indices = sorted([int(feature.split("_")[-1]) for feature in top_features])

    # Filter data to only include top features
    train_data = train_data[:, top_feature_indices]

    return train_data, train_label, classify_dim


##############################################

def process_data_annotation_query(args, feature_dir, split_folder):
    if args.adt != "NULL" and args.atac != "NULL" and args.rna != "NULL":
        mode = "scRNA+scADT+scATAC"

        rna_data_path = args.rna
        adt_data_path = args.adt
        atac_data_path = args.atac

        train_label = torch.load(args.balanced_label)

        # Load data
        rna_data = read_h5_data(rna_data_path)
        adt_data = read_h5_data(adt_data_path)
        atac_data = read_h5_data(atac_data_path)

        test_data = torch.cat((rna_data, adt_data, atac_data), 1)

        classify_dim = (max(train_label)+1).cpu().numpy()

        ### Hybrid Feature Selection Strategy
        # Load the original labels
        df_original = pd.read_csv(args.cty)
        cell_type_counts = df_original['x'].value_counts()

        # Determine the threshold for rare and major classes
        median_count = cell_type_counts.median()
        rare_cell_types = cell_type_counts[cell_type_counts < median_count].index.tolist()

        top_features = []

        # Process each feature file
        for filename in os.listdir(feature_dir):
            if filename.endswith('.csv'):
                filepath = os.path.join(feature_dir, filename)
                
                # Read the CSV file
                df = pd.read_csv(filepath)
                
                # Get cell type from the filename
                cell_type = re.search('fs\.(.*?)_', filename).group(1)
                
                # number of top features to extract
                if cell_type in rare_cell_types:
                    num_features = 100  # number of top features for minor cell types
                else:
                    num_features = 100  # number of top features for major cell types
                
                # Get the top features
                top_features_file = df['Feature Name'].head(num_features).tolist()
                
                # Add these features to the list
                top_features.extend(top_features_file)

        # Get the unique set of top features across all files
        top_features = list(set(top_features))

        # Extract indices from the features and sort them
        top_feature_indices = sorted([int(feature.split("_")[-1]) for feature in top_features])

        # Filter data to only include top features
        test_data = test_data[:, top_feature_indices]

    if args.adt != "NULL" and args.atac == "NULL":
        mode = "scRNA+scADT"

        rna_data_path = args.rna
        adt_data_path = args.adt

        train_label = torch.load(args.balanced_label)

        # Load data
        rna_data = read_h5_data(rna_data_path)
        adt_data = read_h5_data(adt_data_path)

        test_data = torch.cat((rna_data, adt_data), 1)    

        classify_dim = (max(train_label)+1).cpu().numpy()

        ### Hybrid Feature Selection Strategy
        # Load the original labels
        df_original = pd.read_csv(args.cty)
        cell_type_counts = df_original['x'].value_counts()

        # Determine the threshold for rare and major classes
        median_count = cell_type_counts.median()
        rare_cell_types = cell_type_counts[cell_type_counts < median_count].index.tolist()

        top_features = []

        # Process each feature file
        for filename in os.listdir(feature_dir):
            if filename.endswith('.csv'):
                filepath = os.path.join(feature_dir, filename)
                
                # Read the CSV file
                df = pd.read_csv(filepath)
                
                # Get cell type from the filename
                cell_type = re.search('fs\.(.*?)_', filename).group(1)
                
                # number of top features to extract
                if cell_type in rare_cell_types:
                    num_features = 100  # number of top features for minor cell types
                else:
                    num_features = 100  # number of top features for major cell types
                
                # Get the top features
                top_features_file = df['Feature Name'].head(num_features).tolist()
                
                # Add these features to the list
                top_features.extend(top_features_file)

        # Get the unique set of top features across all files
        top_features = list(set(top_features))

        # Extract indices from the features and sort them
        top_feature_indices = sorted([int(feature.split("_")[-1]) for feature in top_features])

        # Filter data to only include top features
        test_data = test_data[:, top_feature_indices]

    if args.adt == "NULL" and args.atac != "NULL":
        mode = "scRNA+scATAC"

        rna_data_path = args.rna
        atac_data_path = args.atac

        train_label = torch.load(args.balanced_label)

        # Load data
        rna_data = read_h5_data(rna_data_path)
        atac_data = read_h5_data(atac_data_path)

        test_data = torch.cat((rna_data, atac_data), 1)    

        classify_dim = (max(train_label)+1).cpu().numpy()

        ### Hybrid Feature Selection Strategy
        # Load the original labels
        df_original = pd.read_csv(args.cty)
        cell_type_counts = df_original['x'].value_counts()

        # Determine the threshold for rare and major classes
        median_count = cell_type_counts.median()
        rare_cell_types = cell_type_counts[cell_type_counts < median_count].index.tolist()

        top_features = []

        # Process each feature file
        for filename in os.listdir(feature_dir):
            if filename.endswith('.csv'):
                filepath = os.path.join(feature_dir, filename)
                
                # Read the CSV file
                df = pd.read_csv(filepath)
                
                # Get cell type from the filename
                cell_type = re.search('fs\.(.*?)_', filename).group(1)
                
                # number of top features to extract
                if cell_type in rare_cell_types:
                    num_features = 100  # number of top features for minor cell types
                else:
                    num_features = 100  # number of top features for major cell types
                
                # Get the top features
                top_features_file = df['Feature Name'].head(num_features).tolist()
                
                # Add these features to the list
                top_features.extend(top_features_file)

        # Get the unique set of top features across all files
        top_features = list(set(top_features))

        # Extract indices from the features and sort them
        top_feature_indices = sorted([int(feature.split("_")[-1]) for feature in top_features])

        # Filter data to only include top features
        test_data = test_data[:, top_feature_indices]

    if args.adt == "NULL" and args.atac == "NULL":

        mode = "scRNA-seq"

        rna_data_path = args.rna

        train_label = torch.load(args.balanced_label)
        
        test_data = read_h5_data(rna_data_path)

        classify_dim = (max(train_label)+1).cpu().numpy()

        ### Hybrid Feature Selection Strategy
        # Load the original labels
        df_original = pd.read_csv(args.cty)
        cell_type_counts = df_original['x'].value_counts()

        # Determine the threshold for rare and major classes
        median_count = cell_type_counts.median()
        rare_cell_types = cell_type_counts[cell_type_counts < median_count].index.tolist()

        top_features = []
        
        # Process each feature file
        for filename in os.listdir(feature_dir):
            if filename.endswith('.csv'):
                filepath = os.path.join(feature_dir, filename)
                
                # Read the CSV file
                df = pd.read_csv(filepath)
                
                # Get cell type from the filename
                cell_type = re.search('fs\.(.*?)_', filename).group(1)
                
                # number of top features to select
                if cell_type in rare_cell_types:
                    num_features = 100  # number of top features for rare cell types
                else:
                    num_features = 100  # number of top features for major cell types
                
                # Get the top features
                top_features_file = df['Feature Name'].head(num_features).tolist()
                
                # Add these features to the list
                for feature in top_features_file:
                    if feature not in top_features:
                        top_features.append(feature)

        # Get the unique set of top features across all files
        top_features= list(set(top_features))

        # Extract indices from the features and sort them
        top_feature_indices = sorted([int(feature.split("_")[-1]) for feature in top_features])

        # Filter data to only include top features
        test_data = test_data[:, top_feature_indices]
    
    if args.rna == "NULL" and args.atac == "NULL":

        mode = "scADT-seq"

        adt_data_path = args.adt

        train_label = torch.load(args.balanced_label)
    
        test_data = read_h5_data(adt_data_path)

        classify_dim = (max(train_label)+1).cpu().numpy()

        ### Hybrid Feature Selection Strategy
        # Load the original labels
        df_original = pd.read_csv(args.cty)
        cell_type_counts = df_original['x'].value_counts()

        # Determine the threshold for rare and major classes
        median_count = cell_type_counts.median()
        rare_cell_types = cell_type_counts[cell_type_counts < median_count].index.tolist()

        top_features = []
        
        # Process each feature file
        for filename in os.listdir(feature_dir):
            if filename.endswith('.csv'):
                filepath = os.path.join(feature_dir, filename)
                
                # Read the CSV file
                df = pd.read_csv(filepath)
                
                # Get cell type from the filename
                cell_type = re.search('fs\.(.*?)_', filename).group(1)
                
                # number of top features to select
                if cell_type in rare_cell_types:
                    num_features = 100  # number of top features for rare cell types
                else:
                    num_features = 100  # number of top features for major cell types
                
                # Get the top features
                top_features_file = df['Feature Name'].head(num_features).tolist()
                
                # Add these features to the list
                for feature in top_features_file:
                    if feature not in top_features:
                        top_features.append(feature)

        # Get the unique set of top features across all files
        top_features= list(set(top_features))

        # Extract indices from the features and sort them
        top_feature_indices = sorted([int(feature.split("_")[-1]) for feature in top_features])

        # Filter data to only include top features
        test_data = test_data[:, top_feature_indices]

    if args.rna == "NULL" and args.adt == "NULL":

        mode = "scATAC-seq"

        atac_data_path = args.atac

        train_label = torch.load(args.balanced_label)
    
        test_data = read_h5_data(atac_data_path)

        classify_dim = (max(train_label)+1).cpu().numpy()

        ### Hybrid Feature Selection Strategy
        # Load the original labels
        df_original = pd.read_csv(args.cty)
        cell_type_counts = df_original['x'].value_counts()

        # Determine the threshold for rare and major classes
        median_count = cell_type_counts.median()
        rare_cell_types = cell_type_counts[cell_type_counts < median_count].index.tolist()

        top_features = []
        
        # Process each feature file
        for filename in os.listdir(feature_dir):
            if filename.endswith('.csv'):
                filepath = os.path.join(feature_dir, filename)
                
                # Read the CSV file
                df = pd.read_csv(filepath)
                
                # Get cell type from the filename
                cell_type = re.search('fs\.(.*?)_', filename).group(1)
                
                # number of top features to select
                if cell_type in rare_cell_types:
                    num_features = 100  # number of top features for rare cell types
                else:
                    num_features = 100  # number of top features for major cell types
                
                # Get the top features
                top_features_file = df['Feature Name'].head(num_features).tolist()
                
                # Add these features to the list
                for feature in top_features_file:
                    if feature not in top_features:
                        top_features.append(feature)

        # Get the unique set of top features across all files
        top_features= list(set(top_features))

        # Extract indices from the features and sort them
        top_feature_indices = sorted([int(feature.split("_")[-1]) for feature in top_features])

        # Filter data to only include top features
        test_data = test_data[:, top_feature_indices]


    return test_data, train_label, classify_dim


##############################################

class Mish(nn.Module):
    def forward(self, x):
        return x * torch.tanh(F.softplus(x))


##############################################

class NN_Classifier(nn.Module):
    def __init__(self, feature_num, num_classes):
        super(NN_Classifier, self).__init__()
        self.net = nn.Sequential(
            nn.Linear(feature_num, 128), 
            Mish(),                        
            nn.Linear(128, num_classes)
        )
    def forward(self, x):
        return self.net(x)


##############################################

def accuracy(output, target, topk=(1,)):
    with torch.no_grad():
        target = target.to(device)

        maxk = max(topk)
        batch_size = target.size(0)

        _, pred = output.topk(maxk, 1, True, True)
        pred = pred.t()
        correct = pred.eq(target.view(1, -1).expand_as(pred))

        res = []
        for k in topk:
            correct_k = correct[:k].view(-1).float().sum(0, keepdim=True)
            res.append(correct_k.mul_(100.0 / batch_size))
            
        return res


##################################################

def clip_counts(new_data, new_label, median_index_length):
    unique_labels = torch.unique(new_label)
    
    indices_to_keep = []
    
    for label in unique_labels:
        indices_of_samples = torch.nonzero(new_label == label, as_tuple=True)[0]
        
        if len(indices_of_samples) > median_index_length:
            # Randomly select samples to keep
            indices_to_keep += indices_of_samples[torch.randperm(len(indices_of_samples))[:median_index_length]].tolist()
        else:
            indices_to_keep += indices_of_samples.tolist()
            
    # Select desired samples from new_data and new_label
    new_data = new_data[indices_to_keep]
    new_label = new_label[indices_to_keep]
    
    return new_data, new_label


##################################################

def perform_data_augmentation(label_to_name_mapping, train_num, classify_dim, train_label, train_data, model, args):

    # Create an empty list to store tuples of class index and the number of samples in that class
    stage1_list = []

    new_label_names = []

    # Iterate through each class index and append a tuple to the list
    for i in np.arange(0, classify_dim):
        stage1_list.append([i, train_num[i]])
    stage1_df = pd.DataFrame(stage1_list)

    # Create a copy of stage1_df for adjustments
    adjusted_df = stage1_df.copy()

    # Calculate the median number of samples for classes
    sorted_train_num = np.sort(train_num)

    if classify_dim % 2 == 0:
        mid_idx = int(classify_dim / 2)
        middle_1 = sorted_train_num[mid_idx - 1]
        middle_2 = sorted_train_num[mid_idx]
        
        # Calculate the average of the two middle values
        train_median = (middle_1 + middle_2) / 2
        
        # Update the class with the largest of the two middle values to the average
        larger_mid_class = stage1_df[stage1_df[1] == max(middle_1, middle_2)][0].iloc[0]
        adjusted_df.loc[adjusted_df[0] == larger_mid_class, 1] = train_median

    else:
        train_median = np.median(train_num)

    # split classes relative to the median
    train_major = adjusted_df[adjusted_df[1] > train_median]
    train_minor = adjusted_df[adjusted_df[1] < train_median]

    anchor_fold = np.array(train_median / train_minor[1])  
    minor_cts   = train_minor[0].to_numpy()                
    major_cts   = train_major[0].to_numpy()           

    median_classes = adjusted_df.loc[adjusted_df[1] == train_median, 0].tolist()

    median_ct      = median_classes[0]
    median_index   = (train_label == median_ct).nonzero(as_tuple=True)[0]

    if classify_dim % 2 == 0 and len(median_index) > train_median:
        median_index = median_index[torch.randperm(len(median_index))[: int(train_median)]]

    extra_indices = torch.cat([
        (train_label == ct).nonzero(as_tuple=True)[0]
        for ct in median_classes[1:]
    ]) if len(median_classes) > 1 else torch.tensor([], dtype=torch.long)

    anchor_data  = train_data [median_index]
    anchor_label = train_label[median_index]

    new_data  = anchor_data.to(device)
    new_label = anchor_label.to(device)
    new_label_names.extend([label_to_name_mapping[median_ct][0]] * len(anchor_data))

    if extra_indices.numel() > 0:
        extra_data  = train_data [extra_indices]
        extra_label = train_label[extra_indices]
        new_data    = torch.cat([new_data,  extra_data.to(device)], 0)
        new_label   = torch.cat([new_label, extra_label.to(device)], 0)
        for ct in median_classes[1:]:
            ct_count = int((extra_label == ct).sum().item())
            new_label_names.extend([label_to_name_mapping[ct][0]] * ct_count)

    ### Randomly downsample the major cell types
    j = 0
    for major_ct in major_cts:

        # Get the number of samples for the current major cell type
        Sample_num = np.array(train_major[1])[j]

        # Create a list of indices from 0 to anchor_num - 1
        N = range(int(Sample_num))

        # Get the indices of the samples belonging to the current major cell type
        index = (train_label == major_ct).nonzero(as_tuple = True)[0]

        # Get the data and labels of the samples belonging to the current major cell type
        Major_data = train_data[index.tolist(), :]
        Major_label = train_label[index.tolist()]

        # Randomly select 'train_median' samples from the current major cell type
        ds_index = random.sample(N, len(median_index))

        # Downsample the data and labels using the random indices
        Major_data = Major_data[ds_index, :]
        Major_label = Major_label[ds_index]

        # Concatenate the downsampled data and labels to the new data and label tensors
        new_data = torch.cat((new_data, Major_data.to(device)), 0)
        new_label = torch.cat((new_label, Major_label.to(device)), 0)

        j = j + 1

        new_label_names.extend([label_to_name_mapping[major_ct][0]] * len(Major_data))


    ### Augment the minor cell types
    j = 0
    for minor_ct in minor_cts:
        # Calculate the augmentation fold and the remaining number of cells to reach train_median
        aug_fold = int((anchor_fold[j]))

        remaining_cell = train_median - aug_fold * np.array(train_minor[1])[j]

        # Get the indices of the samples belonging to the current minor cell type
        index = (train_label == minor_ct).nonzero(as_tuple = True)[0]

        # Get the data and labels of the samples belonging to the current minor cell type
        Minor_data = train_data[index.tolist(), :]
        Minor_label = train_label[index.tolist()]

        # Create a custom dataset and dataloader for the current minor cell type
        Minor_transform_dataset = MyDataset(Minor_data, Minor_label)
        Minor_dl = DataLoader(Minor_transform_dataset, batch_size = args.batch_size, shuffle = True, num_workers = 0, drop_last = False)

        # Generate new data samples using the VAE model
        reconstructed_data, reconstructed_label, real_data = get_vae_simulated_data_from_sampling(model, Minor_dl)

        # Clip the reconstructed_data values to the range of real_data values
        reconstructed_data[reconstructed_data > torch.max(real_data)] = torch.max(real_data)
        reconstructed_data[reconstructed_data < torch.min(real_data)] = torch.min(real_data)
        reconstructed_data[torch.isnan(reconstructed_data)] = torch.max(real_data)

        # Concatenate the generated data and labels to the new data and label tensors
        new_data = torch.cat((new_data, reconstructed_data.to(device)), 0)
        new_label = torch.cat((new_label, reconstructed_label.to(device)), 0)

        new_label_names.extend([label_to_name_mapping[minor_ct][0]] * len(reconstructed_data))

        # Repeat the data generation process for the remaining augmentation folds
        for i in range(aug_fold - 1):
            # Generate new data samples using the VAE model
            reconstructed_data, reconstructed_label, real_data = get_vae_simulated_data_from_sampling(model, Minor_dl)

            # Clip the reconstructed_data values to the range of real_data values
            reconstructed_data[reconstructed_data > torch.max(real_data)] = torch.max(real_data)
            reconstructed_data[reconstructed_data < torch.min(real_data)] = torch.min(real_data)
            reconstructed_data[torch.isnan(reconstructed_data)] = torch.max(real_data)

            # Concatenate the generated data and labels to the new data and label tensors
            new_data = torch.cat((new_data, reconstructed_data.to(device)), 0)
            new_label = torch.cat((new_label, reconstructed_label.to(device)), 0)

            new_label_names.extend([label_to_name_mapping[minor_ct][0]] * len(reconstructed_data))

        # Generate new data samples for the remaining cells
        reconstructed_data, reconstructed_label, real_data = get_vae_simulated_data_from_sampling(model, Minor_dl)

        # Clip the reconstructed_data values to the range of real_data values
        reconstructed_data[reconstructed_data > torch.max(real_data)] = torch.max(real_data)
        reconstructed_data[reconstructed_data < torch.min(real_data)] = torch.min(real_data)
        reconstructed_data[torch.isnan(reconstructed_data)] = torch.max(real_data)

        # Add the remaining cells
        # Create a list of indices from 0 to the number of samples in the current minor cell type - 1
        N = range(int(np.array(train_minor[1])[j]))

        # Randomly select the remaining cells to reach train_median
        ds_index = random.sample(N, int(remaining_cell))

        # Downsample the reconstructed_data and reconstructed_label using the random indices
        reconstructed_data = reconstructed_data[ds_index, :]
        reconstructed_label = reconstructed_label[ds_index]

        # Concatenate the downsampled data and labels to the new data and label tensors
        new_data = torch.cat((new_data, reconstructed_data.to(device)), 0)
        new_label = torch.cat((new_label, reconstructed_label.to(device)), 0)

        # Final check
        curr_cell_count = (new_label == minor_ct).sum().item()
        deficit = int(len(median_index) - curr_cell_count)

        if deficit > 0:
            # Generate new data samples using the VAE model
            reconstructed_data, reconstructed_label, real_data = get_vae_simulated_data_from_sampling(model, Minor_dl)

            # Clip the reconstructed_data values to the range of real_data values
            reconstructed_data[reconstructed_data > torch.max(real_data)] = torch.max(real_data)
            reconstructed_data[reconstructed_data < torch.min(real_data)] = torch.min(real_data)
            reconstructed_data[torch.isnan(reconstructed_data)] = torch.max(real_data)

             # Concatenate the generated data and labels to the new data and label tensors
            new_data = torch.cat((new_data, reconstructed_data.to(device)), 0)
            new_label = torch.cat((new_label, reconstructed_label.to(device)), 0)

        new_label_names.extend([label_to_name_mapping[minor_ct][0]] * len(reconstructed_data))

        j = j + 1

    new_data, new_label = clip_counts(new_data, new_label, len(median_index))

    return new_data, new_label, new_label_names
